from __future__ import annotations

from random import randint
from typing import Dict, Generic, Tuple, TypeVar, Union

from gpiozero import Button

from .. import EngineState
from ..pdi.base3_buffer import Base3Buffer
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope, ControlType
from ..protocol.multibyte.multibyte_constants import TMCC2EffectsControl
from ..protocol.sequence.abs_speed_rpm import AbsoluteSpeedRpm
from ..protocol.sequence.labor_effect import LaborEffectDownReq, LaborEffectUpReq
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum, TMCC1HaltCommandEnum
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum

P = TypeVar("P", bound=Union[int, str, Tuple[int], Tuple[int, int], Tuple[int, int, int]])
R = TypeVar("R", bound=CommandReq)


class EngineController(Generic[R]):
    def __init__(
        self,
        speed_pin_1: P = None,
        speed_pin_2: P = None,
        halt_pin: P = None,
        reset_pin: P = None,
        fwd_pin: P = None,
        rev_pin: P = None,
        front_coupler_pin: P = None,
        rear_coupler_pin: P = None,
        start_up_pin: P = None,
        shutdown_pin: P = None,
        boost_pin: P = None,
        brake_pin: P = None,
        bell_pin: P = None,
        horn_pin: P = None,
        rpm_up_pin: P = None,
        rpm_down_pin: P = None,
        labor_up_pin: P = None,
        labor_down_pin: P = None,
        vol_up_pin: P = None,
        vol_down_pin: P = None,
        smoke_on_pin: P = None,
        smoke_off_pin: P = None,
        tower_dialog_pin: P = None,
        engr_dialog_pin: P = None,
        aux1_pin: P = None,
        aux2_pin: P = None,
        aux3_pin: P = None,
        stop_immediate_pin: P = None,
        i2c_adc_address: int = 0x48,
        train_brake_chn: int = None,
        quilling_horn_chn: int = None,
        cmd_repeat: int = 1,
        base_online_pin: P = None,
        base_offline_pin: P = None,
        base_cathode: bool = True,
        base_ping_freq: int = 5,
        held_threshold: float = 0.5,
        held_frequency: float = 0.1,
    ) -> None:
        from .gpio_handler import GpioHandler, PressedHeldDef

        # initial defaults, use update_engine to modify
        self._engine_specified = False
        self._tmcc_id = 1
        self._control_type = ControlType.LEGACY
        self._scope = CommandScope.ENGINE
        self._repeat = cmd_repeat
        self._state = None
        self._held_threshold = held_threshold
        self._held_frequency = held_frequency
        self._last_tmcc_id = None
        self._last_scope = None

        # define a base watcher, if requested
        if base_online_pin is not None or base_offline_pin:
            from .base_watcher import BaseWatcher

            self._base_watcher = BaseWatcher(
                active_pin=base_online_pin,
                inactive_pin=base_offline_pin,
                cathode=base_cathode,
                delay=base_ping_freq,
            )

        # set up for numeric commands
        self._tmcc1_numeric_cmd = CommandReq(TMCC1EngineCommandEnum.NUMERIC)
        self._tmcc2_numeric_cmd = CommandReq(TMCC2EngineCommandEnum.NUMERIC)
        self._numeric_cmd = None

        # the Halt command only exists in TMCC1 form, and it doesn't take an engine address modifier
        if halt_pin is not None:
            self._halt_btn = GpioHandler.make_button(halt_pin)
            self._halt_btn.when_pressed = CommandReq(TMCC1HaltCommandEnum.HALT).as_action(repeat=2)
        else:
            self._halt_btn = None

        # construct the commands; make both the TMCC1 and Legacy versions
        self._tmcc1_speed_cmd: R | None = None
        self._tmcc2_speed_cmd: R | None = None
        self._tmcc1_when_pushed: Dict[Button, R | None] = {}
        self._tmcc2_when_pushed: Dict[Button, R | None] = {}
        self._tmcc1_when_held: Dict[Button, R | None] = {}
        self._tmcc2_when_held: Dict[Button, R | None] = {}
        self._tmcc1_when_pushed_or_held = {}
        self._tmcc2_when_pushed_or_held = {}

        if speed_pin_1 is not None and speed_pin_2 is not None:
            from .py_rotary_encoder import PyRotaryEncoder

            ramp = {
                20: 3,
                50: 2,
                100: 1,
                200: 1,
                300: 1,
                500: 1,
            }
            self._speed_re = PyRotaryEncoder(speed_pin_1, speed_pin_2, wrap=False, ramp=ramp, max_steps=100)
            self._tmcc1_speed_cmd = CommandReq(TMCC1EngineCommandEnum.ABSOLUTE_SPEED)
            self._tmcc2_speed_cmd = AbsoluteSpeedRpm()
        else:
            self._speed_re = None

        if reset_pin is not None:
            self._reset_btn = GpioHandler.make_button(reset_pin)
            self._tmcc1_when_pushed_or_held[self._reset_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.RESET),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
            self._tmcc2_when_pushed_or_held[self._reset_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.RESET),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
        else:
            self._reset_btn = None

        # setup for Aux Commands
        if aux1_pin:
            self._aux1_btn = GpioHandler.make_button(aux1_pin)
            self._tmcc2_when_pushed_or_held[self._aux1_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.AUX1_OPTION_ONE),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
            self._tmcc2_when_pushed_or_held[self._aux1_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.AUX1_OPTION_ONE),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
        else:
            self._aux1_btn = None

        if aux2_pin:
            self._aux2_btn = GpioHandler.make_button(aux2_pin)
            self._tmcc1_when_pushed[self._aux2_btn] = CommandReq(TMCC1EngineCommandEnum.AUX2_OPTION_ONE)
            self._tmcc2_when_pushed[self._aux2_btn] = CommandReq(TMCC2EngineCommandEnum.AUX2_OPTION_ONE)
        else:
            self._aux2_btn = None

        if aux3_pin:
            self._aux3_btn = GpioHandler.make_button(aux3_pin)
            self._tmcc1_when_pushed[self._aux3_btn] = CommandReq(TMCC1EngineCommandEnum.AUX3_OPTION_ONE)
            self._tmcc2_when_pushed[self._aux3_btn] = CommandReq(TMCC2EngineCommandEnum.AUX3_OPTION_ONE)
        else:
            self._aux3_btn = None

        if stop_immediate_pin:
            self._stop_btn = GpioHandler.make_button(stop_immediate_pin)
            self._tmcc1_when_pushed[self._stop_btn] = CommandReq(TMCC1EngineCommandEnum.SPEED_STOP_HOLD)
            self._tmcc2_when_pushed[self._stop_btn] = CommandReq(TMCC2EngineCommandEnum.STOP_IMMEDIATE)
        else:
            self._stop_btn = None

        if fwd_pin is not None:
            self._fwd_btn = GpioHandler.make_button(fwd_pin)
            self._tmcc1_when_pushed[self._fwd_btn] = CommandReq(TMCC1EngineCommandEnum.FORWARD_DIRECTION)
            self._tmcc2_when_pushed[self._fwd_btn] = CommandReq(TMCC2EngineCommandEnum.FORWARD_DIRECTION)
        else:
            self._fwd_btn = None

        if rev_pin is not None:
            self._rev_btn = GpioHandler.make_button(rev_pin)
            self._tmcc1_when_pushed[self._rev_btn] = CommandReq(TMCC1EngineCommandEnum.REVERSE_DIRECTION)
            self._tmcc2_when_pushed[self._rev_btn] = CommandReq(TMCC2EngineCommandEnum.REVERSE_DIRECTION)
        else:
            self._rev_btn = None

        if front_coupler_pin is not None:
            self._fwd_cplr_btn = GpioHandler.make_button(front_coupler_pin)
            self._tmcc1_when_pushed[self._fwd_cplr_btn] = CommandReq(TMCC1EngineCommandEnum.FRONT_COUPLER)
            self._tmcc2_when_pushed[self._fwd_cplr_btn] = CommandReq(TMCC2EngineCommandEnum.FRONT_COUPLER)
        else:
            self._fwd_cplr_btn = None

        if rear_coupler_pin is not None:
            self._rev_cplr_btn = GpioHandler.make_button(rear_coupler_pin)
            self._tmcc1_when_pushed[self._rev_cplr_btn] = CommandReq(TMCC1EngineCommandEnum.REAR_COUPLER)
            self._tmcc2_when_pushed[self._rev_cplr_btn] = CommandReq(TMCC2EngineCommandEnum.REAR_COUPLER)
        else:
            self._rev_cplr_btn = None

        if start_up_pin is not None:
            self._start_up_btn = GpioHandler.make_button(start_up_pin)
            self._tmcc1_when_pushed[self._start_up_btn] = None
            self._tmcc2_when_pushed_or_held[self._start_up_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.START_UP_IMMEDIATE),
                CommandReq(TMCC2EngineCommandEnum.START_UP_DELAYED),
                held_threshold=self._held_threshold,
            )
        else:
            self._start_up_btn = None

        if shutdown_pin is not None:
            self._shutdown_btn = GpioHandler.make_button(shutdown_pin)
            self._tmcc1_when_pushed[self._shutdown_btn] = CommandReq(TMCC1EngineCommandEnum.SHUTDOWN_IMMEDIATE)
            self._tmcc2_when_pushed_or_held[self._shutdown_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.SHUTDOWN_IMMEDIATE),
                CommandReq(TMCC2EngineCommandEnum.SHUTDOWN_DELAYED),
                held_threshold=self._held_threshold,
            )
        else:
            self._shutdown_btn = None

        if bell_pin is not None:
            self._bell_btn = GpioHandler.make_button(bell_pin)
            self._tmcc1_when_pushed_or_held[self._bell_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.RING_BELL),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
            self._tmcc2_when_pushed_or_held[self._bell_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.BELL_ONE_SHOT_DING, data=3),
                CommandReq(TMCC2EngineCommandEnum.RING_BELL),
                held_threshold=1.5,
            )
        else:
            self.bell_pin = None

        if horn_pin is not None:
            self._horn_btn = GpioHandler.make_button(horn_pin)
            self._tmcc1_when_pushed_or_held[self._horn_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.BLOW_HORN_ONE),
                repeat_action=5,
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=0.05,
            )
            self._tmcc2_when_pushed_or_held[self._horn_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.BLOW_HORN_ONE),
                CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, data=7),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=0.05,
                data=lambda b: randint(0, 15),
            )
        else:
            self._horn_btn = None

        if boost_pin is not None:
            self._boost_btn = GpioHandler.make_button(boost_pin)
            self._tmcc1_when_pushed_or_held[self._boost_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.BOOST_SPEED),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
            self._tmcc2_when_pushed_or_held[self._boost_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.BOOST_SPEED),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
        else:
            self._boost_btn = None

        if brake_pin is not None:
            self._brake_btn = GpioHandler.make_button(brake_pin)
            self._tmcc1_when_pushed_or_held[self._brake_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.BRAKE_SPEED),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
            self._tmcc2_when_pushed_or_held[self._brake_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.BRAKE_SPEED),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_frequency,
            )
        else:
            self._brake_btn = None

        if rpm_up_pin is not None:
            self._rpm_up_btn = GpioHandler.make_button(rpm_up_pin)
            self._tmcc1_when_pushed_or_held[self._rpm_up_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.RPM_UP),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
            self._tmcc2_when_pushed_or_held[self._rpm_up_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.RPM_UP),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
        else:
            self._rpm_up_btn = None

        if rpm_down_pin is not None:
            self._rpm_down_btn = GpioHandler.make_button(rpm_down_pin)
            self._tmcc1_when_pushed_or_held[self._rpm_down_btn] = PressedHeldDef(
                CommandReq(TMCC1EngineCommandEnum.RPM_DOWN),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
            self._tmcc2_when_pushed_or_held[self._rpm_down_btn] = PressedHeldDef(
                CommandReq(TMCC2EngineCommandEnum.RPM_DOWN),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
        else:
            self._rpm_down_btn = None

        if labor_up_pin is not None:
            self._labor_up_btn = GpioHandler.make_button(labor_up_pin)
            self._tmcc1_when_pushed[self._labor_up_btn] = None
            self._tmcc2_when_pushed_or_held[self._labor_up_btn] = PressedHeldDef(
                LaborEffectUpReq(),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
        else:
            self._labor_up_btn = None

        if labor_down_pin is not None:
            self._labor_down_btn = GpioHandler.make_button(labor_down_pin)
            self._tmcc1_when_pushed[self._labor_down_btn] = None
            self._tmcc2_when_pushed_or_held[self._labor_down_btn] = PressedHeldDef(
                LaborEffectDownReq(),
                held_threshold=self._held_threshold,
                hold_repeat=True,
                frequency=self._held_threshold,
            )
        else:
            self._labor_down_btn = None

        if vol_up_pin is not None:
            self._vol_up_btn = GpioHandler.make_button(vol_up_pin)
            self._tmcc1_when_pushed[self._vol_up_btn] = CommandReq(TMCC1EngineCommandEnum.VOLUME_UP)
            self._tmcc2_when_pushed[self._vol_up_btn] = CommandReq(TMCC2EngineCommandEnum.VOLUME_UP)
        else:
            self._vol_up_btn = None

        if vol_down_pin is not None:
            self._vol_down_btn = GpioHandler.make_button(vol_down_pin)
            self._tmcc1_when_pushed[self._vol_down_btn] = CommandReq(TMCC1EngineCommandEnum.VOLUME_DOWN)
            self._tmcc2_when_pushed[self._vol_down_btn] = CommandReq(TMCC2EngineCommandEnum.VOLUME_DOWN)
        else:
            self._vol_down_btn = None

        if smoke_on_pin is not None:
            self._smoke_on_btn = GpioHandler.make_button(smoke_on_pin)
            self._tmcc1_when_pushed[self._smoke_on_btn] = CommandReq(TMCC1EngineCommandEnum.SMOKE_ON)
            self._tmcc2_when_pushed[self._smoke_on_btn] = CommandReq(TMCC2EffectsControl.SMOKE_MEDIUM)
        else:
            self._smoke_on_btn = None

        if smoke_off_pin is not None:
            self._smoke_off_btn = GpioHandler.make_button(smoke_off_pin)
            self._tmcc1_when_pushed[self._smoke_off_btn] = CommandReq(TMCC1EngineCommandEnum.SMOKE_OFF)
            self._tmcc2_when_pushed[self._smoke_off_btn] = CommandReq(TMCC2EffectsControl.SMOKE_OFF)
        else:
            self._smoke_off_btn = None

        if tower_dialog_pin is not None:
            self.tower_dialog_btn = GpioHandler.make_button(tower_dialog_pin)
            self._tmcc1_when_pushed[self.tower_dialog_btn] = CommandReq(TMCC1EngineCommandEnum.NUMERIC, data=7)
            self._tmcc2_when_pushed[self.tower_dialog_btn] = CommandReq(TMCC2EngineCommandEnum.NUMERIC, data=7)
        else:
            self.tower_dialog_btn = None

        if engr_dialog_pin is not None:
            self._engr_dialog_btn = GpioHandler.make_button(engr_dialog_pin)
            self._tmcc1_when_pushed[self._engr_dialog_btn] = CommandReq(TMCC1EngineCommandEnum.NUMERIC, data=2)
            self._tmcc2_when_pushed[self._engr_dialog_btn] = CommandReq(TMCC2EngineCommandEnum.NUMERIC, data=2)
        else:
            self._engr_dialog_btn = None

        if quilling_horn_chn is not None:
            from ..gpio.i2c.analog_handler_i2c import QuillingHorn

            self._quilling_horn_cmd = QuillingHorn(channel=quilling_horn_chn, i2c_address=i2c_adc_address)
        else:
            self._quilling_horn_cmd = None

        if train_brake_chn is not None:
            from ..gpio.i2c.analog_handler_i2c import TrainBrake

            self._train_brake_cmd = TrainBrake(channel=train_brake_chn, i2c_address=i2c_adc_address)
        else:
            self._train_brake_cmd = None

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @property
    def control_type(self) -> ControlType:
        return self._control_type

    @property
    def is_legacy(self) -> bool:
        return self._control_type and self._control_type.is_legacy

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @property
    def halt_btn(self) -> Button:
        return self._halt_btn

    @property
    def reset_btn(self) -> Button:
        return self._reset_btn

    @property
    def fwd_btn(self) -> Button:
        return self._fwd_btn

    @property
    def rev_btn(self) -> Button:
        return self._rev_btn

    @property
    def start_up_btn(self) -> Button:
        return self._start_up_btn

    @property
    def shutdown_btn(self) -> Button:
        return self._shutdown_btn

    @property
    def bell_btn(self) -> Button:
        return self._bell_btn

    @property
    def horn_btn(self) -> Button:
        return self._horn_btn

    @property
    def boost_btn(self) -> Button:
        return self._boost_btn

    @property
    def brake_btn(self) -> Button:
        return self._brake_btn

    @property
    def rpm_up_btn(self) -> Button:
        return self._rpm_up_btn

    @property
    def rpm_down_btn(self) -> Button:
        return self._rpm_down_btn

    @property
    def labor_up_btn(self) -> Button:
        return self._labor_up_btn

    @property
    def labor_down_btn(self) -> Button:
        return self._labor_down_btn

    @property
    def volume_up_btn(self) -> Button:
        return self._vol_up_btn

    @property
    def volume_down_btn(self) -> Button:
        return self._vol_down_btn

    @property
    def smoke_on_btn(self) -> Button:
        return self._smoke_on_btn

    @property
    def smoke_off_btn(self) -> Button:
        return self._smoke_off_btn

    def update(
        self,
        tmcc_id: int,
        scope: CommandScope = CommandScope.ENGINE,
        state: EngineState = None,
    ) -> None:
        from .gpio_handler import GpioHandler

        """
        When a new engine/train is selected, redo the button bindings to
        reflect the new engine/train tmcc_id
        """
        self._engine_specified = True
        self._last_scope = self._scope
        self._last_tmcc_id = self._tmcc_id
        self._scope = scope
        self._tmcc_id = tmcc_id
        self._state = cur_state = state
        print(f"EngineController.update: scope={scope}, tmcc_id={tmcc_id}, state={state}")
        if cur_state is None or cur_state.control_type is None:
            self._control_type = ControlType.LEGACY
        else:
            self._control_type = ControlType.by_value(cur_state.control_type)
        # request state update from Base, if present
        Base3Buffer.request_state_update(tmcc_id, scope)
        # update buttons
        if self.is_legacy:
            max_speed = 200
            speed_limit = 195
            num_steps = 100
            when_pushed = self._tmcc2_when_pushed
            when_held = self._tmcc2_when_held
            when_pushed_or_held = self._tmcc2_when_pushed_or_held
            speed_cmd = self._tmcc2_speed_cmd
            quilling_horn_cmd = self._quilling_horn_cmd
            train_brake_cmd = self._train_brake_cmd
            numeric_cmd = self._tmcc2_numeric_cmd
        else:
            max_speed = 31
            speed_limit = 27
            num_steps = 32
            when_pushed = self._tmcc1_when_pushed
            when_held = self._tmcc1_when_held
            when_pushed_or_held = self._tmcc1_when_pushed_or_held
            speed_cmd = self._tmcc1_speed_cmd
            quilling_horn_cmd = None
            train_brake_cmd = None
            numeric_cmd = self._tmcc1_numeric_cmd

        # reset the when_pressed button handlers
        for btn, cmd in when_pushed.items():
            if cmd:
                cmd.address = self._tmcc_id
                cmd.scope = scope
                btn.when_pressed = cmd.as_action(repeat=self._repeat)
            else:
                if btn.when_pressed:
                    btn.when_pressed = None

        # reset the when_held button handlers
        for btn, cmd in when_held.items():
            if cmd:
                cmd.address = self._tmcc_id
                cmd.scope = scope
                btn.when_held = cmd.as_action()
            else:
                if btn.when_held:
                    btn.when_held = None

        # reset the when_pushed_or_held button handlers
        for btn, phd in when_pushed_or_held.items():
            if phd:
                phd.update_target(self._tmcc_id, scope=scope)
                btn.when_pressed = phd.as_action()
            else:
                if btn.when_pressed:
                    btn.when_pressed = None

        # reset the rotary encoder handlers
        if speed_cmd:
            speed_cmd.address = self._tmcc_id
            speed_cmd.scope = scope
            max_speed = (
                cur_state.max_speed if cur_state and cur_state.max_speed and cur_state.max_speed != 255 else max_speed
            )
            speed_limit = (
                cur_state.speed_limit
                if cur_state and cur_state.speed_limit and cur_state.speed_limit != 255
                else speed_limit
            )
            max_speed = min(max_speed, speed_limit)
            steps_to_speed = GpioHandler.make_interpolator(max_speed, 0, -num_steps, num_steps)
            speed_to_steps = GpioHandler.make_interpolator(num_steps, -num_steps, 0, max_speed)
            print(speed_cmd)
            self._speed_re.update_action(speed_cmd, cur_state, steps_to_speed, speed_to_steps)

        # reset the quilling horn
        if quilling_horn_cmd:
            quilling_horn_cmd.update_action(self._tmcc_id, scope)

        # reset the train brake
        if train_brake_cmd:
            train_brake_cmd.update_action(self._tmcc_id, scope)

        # reset the numeric command
        if numeric_cmd:
            numeric_cmd.address = self._tmcc_id
            numeric_cmd.scope = scope
            self._numeric_cmd = numeric_cmd
        else:
            self._numeric_cmd = None

    def on_speed_changed(self, new_speed: int) -> None:
        if self._speed_re and new_speed is not None:
            self._speed_re.update_data(new_speed)

    def on_numeric(self, key: str):
        if self._numeric_cmd:
            self._numeric_cmd.as_action()(new_data=int(key))



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import threading


def singleton(cls):
    _instances = {}
    _lock: threading.Lock = threading.Lock()

    def wrapper(*args, **kwargs):
        with _lock:
            if cls not in _instances:
                instance = cls(*args, **kwargs)
                instance._initialized = False
                _instances[cls] = instance
            return _instances[cls]

    return wrapper



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from typing import TypeVar

import sys

from .multibyte_constants import TMCCPrefixEnum, TMCC2ParameterEnum

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from ..command_req import CommandReq
from ..constants import DEFAULT_ADDRESS, CommandScope
from ..tmcc2.tmcc2_constants import TMCC2_SCOPE_TO_FIRST_BYTE_MAP, LEGACY_MULTIBYTE_COMMAND_PREFIX
from .multibyte_constants import TMCC2MultiByteEnum

E = TypeVar("E", bound=TMCC2MultiByteEnum)

MULTIBYTE_PREFIX_BYTE = LEGACY_MULTIBYTE_COMMAND_PREFIX.to_bytes(1, byteorder="big")


class MultiByteReq(CommandReq, ABC):
    __metaclass__ = ABCMeta

    @classmethod
    def build(
        cls, command: TMCC2MultiByteEnum, address: int = DEFAULT_ADDRESS, data: int = 0, scope: CommandScope = None
    ) -> Self:
        if isinstance(command, TMCC2ParameterEnum):
            from .param_command_req import ParameterCommandReq  # noqa: E402

            return ParameterCommandReq.build(command, address, data, scope)

    @classmethod
    def vet_bytes(cls, param: bytes, cmd_type: str = "", raise_exception: bool = True) -> tuple[bool, bool]:
        is_vmb = False
        is_d4 = False

        if not param:
            if raise_exception:
                raise ValueError("Command requires at least 9 bytes")
            else:
                return is_vmb, is_d4
        if len(param) < 9:
            if raise_exception:
                raise ValueError(f"{cmd_type} command requires at least 9 bytes {param.hex(':')}")
            else:
                return is_vmb, is_d4
        if (
            len(param) == 9
            and param[3] == LEGACY_MULTIBYTE_COMMAND_PREFIX
            and param[6] == LEGACY_MULTIBYTE_COMMAND_PREFIX
        ):
            is_vmb = True
        elif (
            len(param) == 21
            and param[7] == LEGACY_MULTIBYTE_COMMAND_PREFIX
            and param[14] == LEGACY_MULTIBYTE_COMMAND_PREFIX
        ):
            is_vmb = is_d4 = True
        return is_vmb, is_d4

    @classmethod
    def from_bytes(cls, param: bytes, from_tmcc_rx: bool = False, is_tmcc4: bool = False) -> Self:
        is_vmb, _ = cls.vet_bytes(param, "Multy-byte")
        if is_vmb is True:
            index = 0x00F0 & int.from_bytes(param[1:3], byteorder="big")
            prefix = TMCCPrefixEnum.by_value(index)
            if prefix == TMCCPrefixEnum.PARAMETER:
                from .param_command_req import ParameterCommandReq  # noqa: E402

                return ParameterCommandReq.from_bytes(param, from_tmcc_rx, is_tmcc4)
            elif prefix == TMCCPrefixEnum.R4LC:
                from .r4lc_command_req import R4LCCommandReq  # noqa: E402

                return R4LCCommandReq.from_bytes(param, from_tmcc_rx, is_tmcc4)
            elif prefix == TMCCPrefixEnum.VARIABLE:
                from .dcds_command_req import VariableCommandReq

                return VariableCommandReq.from_bytes(param, from_tmcc_rx, is_tmcc4)
        raise ValueError(f"Invalid multibyte command: : {param.hex(':')}")

    def __init__(
        self,
        command_def_enum: E,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
    ) -> None:
        super().__init__(command_def_enum, address, data, scope)

    @property
    def num_bytes(self) -> int:
        """
        Returns the number of bytes in this command. Except for the
        Variable length  multibyte commands, this is always 9 bytes
        in three sets of three bytes a piece.
        """
        if 1 <= self.address <= 99:
            return 9
        else:  # 4-digit addressing
            return 21

    @property
    def as_bytes(self) -> bytes:
        byte_str = bytes()
        byte_str += TMCC2_SCOPE_TO_FIRST_BYTE_MAP[self.scope].to_bytes(1, byteorder="big")
        byte_str += self._word_1
        byte_str += self.word_prefix
        byte_str += self._word_2
        byte_str += self.word_prefix
        byte_str += self.checksum(byte_str)
        if self.address > 99:
            # handle 4-digit address
            address_bytes = str(self.address).zfill(4).encode()
            tmp = bytes()
            for i in range(0, len(byte_str), 3):
                tmp += byte_str[i : i + 3] + address_bytes
            byte_str = tmp
        return byte_str

    @property
    def multibyte_word_prefix(self):
        return LEGACY_MULTIBYTE_COMMAND_PREFIX.to_bytes(1, byteorder="big")

    @property
    def address_byte(self) -> bytes:
        e_t = 1 if self.scope == CommandScope.TRAIN else 0
        if 1 <= self.address <= 99:
            address = self.address
        else:
            address = 0  # prep for 4-digit
        return ((address << 1) + e_t).to_bytes(1, "big")

    @property
    def word_prefix(self) -> bytes:
        return MULTIBYTE_PREFIX_BYTE + self.address_byte

    @property
    def _word_1(self) -> bytes:
        if 1 <= self.address <= 99:
            address = self.address
        else:
            address = 0  # prep for 4-digit
        return ((address << 1) + 1).to_bytes(1, "big") + self.index_byte

    @property
    def _word_2(self) -> bytes:
        return self.data_byte

    # noinspection PyUnusedLocal
    @staticmethod
    def checksum(data: bytes = None) -> bytes:
        """
        Calculate the checksum of a fixed-length lionel tmcc2 multibyte command.
        The checksum is calculated adding together the second 2 bytes of the
        parameter index and parameter data words, and the addr byte of the checksum
        word, and returning the 1's complement of that sum mod 256.

        We make use of self.command_scope to determine if the command directed at
        an engine or train.
        """
        byte_sum = 0
        for b in data:
            if b not in {0xF8, 0xF9, 0xFB}:
                byte_sum += int(b)
        return (~(byte_sum % 256) & 0xFF).to_bytes(1, byteorder="big")  # return 1's complement of sum mod 256

    def _apply_address(self, **kwargs) -> int:
        return self.command_def.bits

    def _apply_data(self, **kwargs) -> int:
        return self.command_def.bits

    @property
    @abstractmethod
    def index_byte(self) -> bytes:
        pass

    @property
    @abstractmethod
    def data_byte(self) -> bytes:
        pass



import abc
from abc import ABC
from ipaddress import IPv4Address, IPv6Address
from typing import Sequence, TypeVar

from ..comm.comm_buffer import CommBuffer
from ..pdi.pdi_req import PdiReq
from ..utils.validations import Validations
from .command_def import CommandDef, CommandDefEnum
from .command_req import CommandReq
from .constants import DEFAULT_BAUDRATE, DEFAULT_PORT, MINIMUM_DURATION_INTERVAL_MSEC, CommandScope

R = TypeVar("R", bound=CommandReq | PdiReq | Sequence[CommandReq | PdiReq])


class CommandBase(ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        command: CommandDefEnum | None,
        command_req: R,
        address: int = 99,
        data: int = 0,
        scope: CommandScope = None,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        self._address = address
        self._command = None  # provided by _build_command method in subclasses
        # validate baudrate
        if baudrate is None or baudrate < 110 or baudrate > 115200 or not isinstance(baudrate, int):
            raise ValueError("baudrate must be between 110 and 115200")
        self._baudrate = baudrate
        # validate port
        if port is None:
            raise ValueError("port cannot be None")
        self._port = port
        # validate server ip address
        self._server, self._port = CommBuffer.parse_server(server, port)

        # persist command information
        self._command_def_enum: CommandDefEnum = command
        self._command_def: CommandDef = command.value if command else None
        self._command_req: R = command_req
        self._data: int = data
        self._scope: CommandScope = scope

        # build_req the command
        self._command = self._build_command()

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @property
    def bits(self) -> int:
        return self._command_req.bits if isinstance(self._command_req, CommandReq) else 0

    @property
    def address(self) -> int:
        return self._address

    @property
    def port(self) -> str:
        return self._port

    @property
    def server(self) -> IPv4Address | IPv6Address | None:
        return self._server

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @property
    def command_bytes(self) -> bytes:
        return self._command

    @property
    def command_prefix(self) -> bytes:
        return self._command_prefix()

    @property
    def command_req(self) -> R:
        return self._command_req

    def send(
        self,
        repeat: int = None,
        delay: float = None,
        duration: float = None,
        interval: int = None,
        shutdown: bool = False,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ):
        """
        Send the command to the LCS SER2 and keep comm buffer alive.
        """
        Validations.validate_int(repeat, min_value=1)
        Validations.validate_int(interval, min_value=MINIMUM_DURATION_INTERVAL_MSEC, allow_none=True)
        Validations.validate_float(delay, min_value=0)
        Validations.validate_float(duration, min_value=0, allow_none=True)
        if isinstance(self.command_req, CommandReq) or isinstance(self.command_req, PdiReq):
            reqs = [self.command_req]
        elif isinstance(self.command_req, Sequence):
            reqs = self.command_req
        else:
            raise ValueError(f"Unknown command type: {type(self.command_req)}")
        for req in reqs:
            req.send(
                repeat=repeat,
                delay=delay,
                duration=duration,
                interval=interval,
                baudrate=baudrate,
                port=port,
                server=server,
            )
        if shutdown:
            buffer = CommBuffer.build(baudrate=baudrate, port=port, server=server)
            buffer.shutdown()
            buffer.join()

    def fire(
        self,
        repeat: int = 1,
        delay: float = 0,
        duration: float = 0,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        """
        Fire the command immediately and shut down comm buffers, once empty
        """
        self.send(
            repeat=repeat,
            delay=delay,
            duration=duration,
            shutdown=True,
            baudrate=baudrate,
            port=port,
            server=server,
        )

    @staticmethod
    def _encode_command(command: int, num_bytes: int = 2) -> bytes:
        return command.to_bytes(num_bytes, byteorder="big")

    @abc.abstractmethod
    def _build_command(self) -> bytes | None:
        return None

    @abc.abstractmethod
    def _command_prefix(self) -> bytes | None:
        return None

    @abc.abstractmethod
    def _encode_address(self, command_op: int) -> bytes | None:
        return None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from .gpio_device import GpioDevice, P
from .state_source import AccessoryStateSource


class CulvertLoader(GpioDevice):
    def __init__(
        self,
        address: int,
        cycle_pin: P,
        cycle_led_pin: P = None,
        command_control: bool = True,
        cathode: bool = True,
    ) -> None:
        if command_control is True:
            cycle_req, self.cycle_btn, self.cycle_led = self.make_button(
                cycle_pin,
                TMCC1AuxCommandEnum.AUX2_OPT_ONE,
                address,
                led_pin=cycle_led_pin,
                cathode=cathode,
            )
            self.cycle_btn.when_pressed = cycle_req.as_action(repeat=2)
        else:
            raise NotImplementedError
        if self.cycle_led:
            self.cache_handler(AccessoryStateSource(address, self.cycle_led, aux2_state=TMCC1AuxCommandEnum.AUX2_ON))


class CulvertUnloader(CulvertLoader):
    pass



from unittest import mock

import builtins
import pytest

from src.pytrain.cli.make_service import MakeService
from ..test_base import TestBase


class TestMakeService(TestBase):
    def test_parser(self):
        # test successful options
        for i in range(2, len("-version")):
            li = ["-version"][:i]
            with pytest.raises(SystemExit) as e:
                MakeService(li)
            assert e.value.code == 0

        with pytest.raises(SystemExit) as e:
            MakeService(["-h"])
        assert e.value.code == 0

        with pytest.raises(SystemExit) as e:
            MakeService(["--help"])
        assert e.value.code == 0

        with mock.patch.object(builtins, "input", return_value="n"):
            # test some positive cases
            assert MakeService("-client".split()) is not None
            assert MakeService("-client -echo".split()) is not None
            assert MakeService("-client -buttons".split()) is not None
            assert MakeService("-client -buttons -start".split()) is not None
            assert MakeService("-server -base -ser2 -echo -buttons_f".split()) is not None

            # test some negative cases
            # neither client nor server specified
            with pytest.raises(SystemExit) as e:
                MakeService("-echo -buttons -start".split())
            assert e.value.code == 2

            # bad arguments
            with pytest.raises(SystemExit) as e:
                MakeService("-service".split())
            assert e.value.code == 2



from __future__ import annotations

from ..pdi.constants import PdiCommand, WiFiAction
from ..pdi.lcs_req import LcsReq

WIFI_MODE_MAP = {0: "AP", 1: "INF", 2: "WPS"}


class WiFiReq(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.WIFI_GET,
        action: WiFiAction = WiFiAction.CONFIG,
        ident: int | None = None,
        error: bool = False,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        if isinstance(data, bytes):
            self._action = WiFiAction(self._action_byte)
        else:
            self._action = action

    @property
    def base_address(self) -> str | None:
        if self._data is not None and self.action == WiFiAction.IP:
            payload = self._data[3:]
            return f"{payload[0]}.{payload[1]}.{payload[2]}.{payload[3]}"
        return None

    @property
    def clients(self) -> list[str]:
        if self._data is not None and self.action == WiFiAction.IP:
            payload = self._data[3:]
            prefix = f"{payload[0]}.{payload[1]}.{payload[2]}."
            the_clients = list()
            for i in range(0, len(payload), 2):
                the_clients.append(f"{prefix}{payload[i + 1]}")
            return the_clients
        return []

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        else:
            payload_bytes = self._data[3:]
            if self.action == WiFiAction.CONNECT:
                return (
                    f"Max Connections: {payload_bytes[0]} Connected: {payload_bytes[1]}"
                    + f" {WIFI_MODE_MAP[payload_bytes[2]]}"
                )
            elif self.action == WiFiAction.IP:
                ip_addr = f"{payload_bytes[0]}.{payload_bytes[1]}.{payload_bytes[2]}.{payload_bytes[3]}"
                payload_bytes = payload_bytes[4:]
                clients = " Clients: "
                for i in range(0, len(payload_bytes), 2):
                    if i > 0:
                        clients += ", "
                    clients += f"{payload_bytes[i + 1]} ({payload_bytes[i]})"
                return f"Base IP: {ip_addr} {clients}"
            elif self.action == WiFiAction.RESPBCASTS:
                return f"Broadcasts {'ENABLED' if payload_bytes[0] == 1 else 'DISABLED'}: {payload_bytes[0]}"
        return super().payload



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import requests


API_KEY = "LionChief-Android-ED2E9A6F5F08"
PROD_INFO_URL = "https://proddb.lionel.com/api/engine/getenginebyhexid/{}"


class ProdInfo:
    @classmethod
    def get_info(cls, bt_id: str) -> dict:
        header = {"LionelApiKey": API_KEY}
        response = requests.get(PROD_INFO_URL.format(bt_id), headers=header)
        if response.status_code == 200:
            return response.json()
        else:
            msg = f"Request for product information on {bt_id} failed with status code {response.status_code}"
            raise requests.RequestException(msg)



import threading
import time
from collections import deque

# noinspection PyPackageRequirements
import pytest

# noinspection PyProtectedMember
from src.pytrain.comm.command_listener import CommandListener, CommandDispatcher
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import DEFAULT_BAUDRATE, DEFAULT_PORT, DEFAULT_QUEUE_SIZE
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1HaltCommandEnum
from src.pytrain.protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum
from test.test_base import TestBase


# noinspection PyTypeChecker
@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmpdir) -> None:
    """
    Fixture to execute asserts before and after a test is run
    """
    # Setup: fill with any logic you want

    yield  # this is where the testing happens

    # Teardown : fill with any logic you want
    if CommandListener.is_built():
        CommandListener().shutdown()
    assert CommandListener.is_built() is False

    if CommandDispatcher.is_built():
        CommandDispatcher().shutdown()
    assert CommandDispatcher.is_built() is False


class TestCommandListener(TestBase):
    def test_command_listener_singleton(self) -> None:
        assert CommandListener.is_built() is False
        listener = CommandListener()
        assert listener.is_built() is True
        assert listener.is_running() is True
        assert isinstance(listener, CommandListener)
        assert listener is CommandListener()
        assert listener.baudrate == DEFAULT_BAUDRATE
        assert listener.port == DEFAULT_PORT
        assert listener.is_alive()
        assert listener.daemon is True
        assert listener._deque is not None  # dequeue should exist
        assert listener._deque.maxlen == DEFAULT_QUEUE_SIZE
        assert isinstance(listener._deque, deque)
        assert not listener._deque  # should be empty
        assert listener._cv is not None
        assert isinstance(listener._cv, threading.Condition)

        # shutdown should clear singleton, forcing a new one to be created
        assert CommandListener._instance is not None
        listener.shutdown()
        assert CommandListener._instance is None
        assert listener.is_built() is False
        assert listener.is_running() is False
        assert CommandListener.is_built() is False
        assert listener != CommandListener()

    def test_command_listener_build(self) -> None:
        listener = CommandListener.build(baudrate=57600)
        assert listener
        assert listener.is_built() is True
        assert listener.is_running() is True
        assert CommandListener.is_built() is True
        assert CommandListener.is_running() is True
        assert listener.baudrate == 57600
        assert listener == CommandListener()
        assert CommandListener(baudrate=9600).baudrate == 57600  # original instance returned

    def test_command_listener_invalid_baudrate(self) -> None:
        with pytest.raises(ValueError, match="Invalid baudrate: 12345"):
            CommandListener.build(baudrate=12345)

    def test_command_listener_run(self) -> None:
        listener = CommandListener.build()
        # add elements to the queue and make sure they appear in deque in the correct order
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 10)
        halt_req = CommandReq.build(TMCC1HaltCommandEnum.HALT)
        with listener._cv:
            listener.offer(ring_req.as_bytes)
            listener.offer(halt_req.as_bytes)
            # deque should contain 6 bytes
            assert len(listener._deque) == 6
            cmd_bytes = ring_req.as_bytes + halt_req.as_bytes
            for i in range(6):
                assert cmd_bytes[i] == listener._deque[i]
        # outside the lock context, consumer threads will run
        time.sleep(0.1)  # allow threads to clear deque
        assert len(listener._deque) == 0  # both entries processed
        # lock should be open too
        assert listener._cv.acquire(blocking=False) is True
        listener._cv.release()



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from queue import Queue
from threading import Thread
from typing import Callable

from .component_state import ComponentState
from ..protocol.constants import PROGRAM_NAME

log = logging.getLogger(__name__)


class StateWatcher(Thread):
    def __init__(self, state: ComponentState, action: Callable) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} State Watcher {state.scope.label} {state.address}")
        self._state = state
        self._action = action
        self._notifier = UpdateNotifier(self)
        self._is_running = True
        self.start()

    def action(self) -> None:
        self._action()

    @property
    def watched(self) -> ComponentState:
        return self._state

    def shutdown(self) -> None:
        self._is_running = False
        if self._notifier:
            self._notifier.shutdown()
        with self._state.synchronizer:
            self._state.synchronizer.notify_all()

    def run(self) -> None:
        while self._state is not None and self._is_running:
            with self._state.synchronizer:
                self._state.synchronizer.wait()
            if self._is_running:
                self._notifier.update_request()


class UpdateNotifier(Thread):
    def __init__(self, watcher: StateWatcher) -> None:
        super().__init__(daemon=True)
        self._watcher = watcher
        self._is_running = True
        self._queue = Queue(64)
        self.start()

    def run(self) -> None:
        while self._is_running:
            # wait for a state change notification if queue is empty
            if self._queue.empty():
                self._queue.get(block=True)
            # clear out queue, the redisplay we're about to trigger covers them
            with self._queue.mutex:
                self._queue.queue.clear()
                self._queue.all_tasks_done.notify_all()
                self._queue.unfinished_tasks = 0
            if self._is_running:
                self._watcher.action()

    def shutdown(self) -> None:
        self._is_running = False
        self.update_request()

    def update_request(self):
        self._queue.put(True)



from __future__ import annotations

from threading import RLock, Thread
from time import sleep
from typing import List

from ..db.component_state_store import ComponentStateStore
from ..db.state_watcher import StateWatcher
from ..gpio.i2c.lcd import LCD_PCF8574_ADDRESS, Lcd
from ..protocol.constants import PROGRAM_NAME, CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum
from .engine_controller import EngineController
from .engine_status import EngineStatus
from .gpio_device import GpioDevice, P
from .keypad import KEYPAD_PCF8574_ADDRESS, Keypad, KeyPadI2C
from ..utils.unique_deque import UniqueDeque

COMMANDS_OF_INTEREST = {
    TMCC1EngineCommandEnum.ABSOLUTE_SPEED,
    TMCC1EngineCommandEnum.FORWARD_DIRECTION,
    TMCC1EngineCommandEnum.REVERSE_DIRECTION,
    TMCC1EngineCommandEnum.NUMERIC,
    TMCC2EngineCommandEnum.ABSOLUTE_SPEED,
    TMCC2EngineCommandEnum.FORWARD_DIRECTION,
    TMCC2EngineCommandEnum.REVERSE_DIRECTION,
    TMCC2EngineCommandEnum.DIESEL_RPM,
    TMCC2EngineCommandEnum.NUMERIC,
    TMCC2EngineCommandEnum.MOMENTUM,
    TMCC2EngineCommandEnum.MOMENTUM_LOW,
    TMCC2EngineCommandEnum.MOMENTUM_MEDIUM,
    TMCC2EngineCommandEnum.MOMENTUM_HIGH,
    TMCC2EngineCommandEnum.TRAIN_BRAKE,
}


class Controller(Thread, GpioDevice):
    @classmethod
    def build(
        cls,
        lcd_address: int = 0x27,
        lcd_rows: int = 4,
        lcd_cols: int = 20,
        is_lcd: bool = False,
        oled_address: int = 0,
        oled_device="ssd1362",
        is_oled: bool = True,
        keypad_address: int | None = KEYPAD_PCF8574_ADDRESS,
        row_pins: List[int | str] = None,
        column_pins: List[int | str] = None,
        base_online_pin: P = None,
        base_offline_pin: P = None,
        base_cathode: bool = True,
        base_ping_freq: int = 5,
        speed_pins: List[int | str] = None,
        halt_pin: P = None,
        reset_pin: P = None,
        fwd_pin: P = None,
        rev_pin: P = None,
        front_coupler_pin: P = None,
        rear_coupler_pin: P = None,
        start_up_pin: P = None,
        shutdown_pin: P = None,
        boost_pin: P = None,
        brake_pin: P = None,
        bell_pin: P = None,
        horn_pin: P = None,
        rpm_up_pin: P = None,
        rpm_down_pin: P = None,
        labor_up_pin: P = None,
        labor_down_pin: P = None,
        vol_up_pin: P = None,
        vol_down_pin: P = None,
        smoke_on_pin: P = None,
        smoke_off_pin: P = None,
        tower_dialog_pin: P = None,
        engr_dialog_pin: P = None,
        aux1_pin: P = None,
        aux2_pin: P = None,
        aux3_pin: P = None,
        stop_immediate_pin: P = None,
        i2c_adc_address: int = 0x48,
        train_brake_chn: int = None,
        quilling_horn_chn: int = None,
        num_lasts: int = 4,
    ) -> Controller:
        if row_pins and column_pins:
            c = Controller(
                lcd_address=lcd_address,
                lcd_rows=lcd_rows,
                lcd_cols=lcd_cols,
                is_lcd=is_lcd,
                oled_address=oled_address,
                oled_device=oled_device,
                is_oled=is_oled,
                row_pins=row_pins,
                column_pins=column_pins,
                base_online_pin=base_online_pin,
                base_offline_pin=base_offline_pin,
                base_cathode=base_cathode,
                base_ping_freq=base_ping_freq,
                speed_pins=speed_pins,
                halt_pin=halt_pin,
                reset_pin=reset_pin,
                fwd_pin=fwd_pin,
                rev_pin=rev_pin,
                front_coupler_pin=front_coupler_pin,
                rear_coupler_pin=rear_coupler_pin,
                start_up_pin=start_up_pin,
                shutdown_pin=shutdown_pin,
                boost_pin=boost_pin,
                brake_pin=brake_pin,
                bell_pin=bell_pin,
                horn_pin=horn_pin,
                rpm_up_pin=rpm_up_pin,
                rpm_down_pin=rpm_down_pin,
                labor_up_pin=labor_up_pin,
                labor_down_pin=labor_down_pin,
                vol_up_pin=vol_up_pin,
                vol_down_pin=vol_down_pin,
                smoke_on_pin=smoke_on_pin,
                smoke_off_pin=smoke_off_pin,
                tower_dialog_pin=tower_dialog_pin,
                engr_dialog_pin=engr_dialog_pin,
                aux1_pin=aux1_pin,
                aux2_pin=aux2_pin,
                aux3_pin=aux3_pin,
                stop_immediate_pin=stop_immediate_pin,
                i2c_adc_address=i2c_adc_address,
                train_brake_chn=train_brake_chn,
                quilling_horn_chn=quilling_horn_chn,
                num_lasts=num_lasts,
            )
        else:
            c = ControllerI2C(
                lcd_address=lcd_address,
                lcd_rows=lcd_rows,
                lcd_cols=lcd_cols,
                is_lcd=is_lcd,
                oled_address=oled_address,
                oled_device=oled_device,
                is_oled=is_oled,
                keypad_address=keypad_address,
                base_online_pin=base_online_pin,
                base_offline_pin=base_offline_pin,
                base_cathode=base_cathode,
                base_ping_freq=base_ping_freq,
                speed_pins=speed_pins,
                halt_pin=halt_pin,
                reset_pin=reset_pin,
                fwd_pin=fwd_pin,
                rev_pin=rev_pin,
                front_coupler_pin=front_coupler_pin,
                rear_coupler_pin=rear_coupler_pin,
                start_up_pin=start_up_pin,
                shutdown_pin=shutdown_pin,
                boost_pin=boost_pin,
                brake_pin=brake_pin,
                bell_pin=bell_pin,
                horn_pin=horn_pin,
                rpm_up_pin=rpm_up_pin,
                rpm_down_pin=rpm_down_pin,
                labor_up_pin=labor_up_pin,
                labor_down_pin=labor_down_pin,
                vol_up_pin=vol_up_pin,
                vol_down_pin=vol_down_pin,
                smoke_on_pin=smoke_on_pin,
                smoke_off_pin=smoke_off_pin,
                tower_dialog_pin=tower_dialog_pin,
                engr_dialog_pin=engr_dialog_pin,
                aux1_pin=aux1_pin,
                aux2_pin=aux2_pin,
                aux3_pin=aux3_pin,
                stop_immediate_pin=stop_immediate_pin,
                i2c_adc_address=i2c_adc_address,
                train_brake_chn=train_brake_chn,
                quilling_horn_chn=quilling_horn_chn,
                num_lasts=num_lasts,
            )
        return c

    def __init__(
        self,
        row_pins: List[P] = None,
        column_pins: List[P] = None,
        speed_pins: List[P] = None,
        halt_pin: P = None,
        reset_pin: P = None,
        fwd_pin: P = None,
        rev_pin: P = None,
        front_coupler_pin: P = None,
        rear_coupler_pin: P = None,
        start_up_pin: P = None,
        shutdown_pin: P = None,
        boost_pin: P = None,
        brake_pin: P = None,
        bell_pin: P = None,
        horn_pin: P = None,
        rpm_up_pin: P = None,
        rpm_down_pin: P = None,
        labor_up_pin: P = None,
        labor_down_pin: P = None,
        vol_up_pin: P = None,
        vol_down_pin: P = None,
        smoke_on_pin: P = None,
        smoke_off_pin: P = None,
        tower_dialog_pin: P = None,
        engr_dialog_pin: P = None,
        aux1_pin: P = None,
        aux2_pin: P = None,
        aux3_pin: P = None,
        stop_immediate_pin: P = None,
        i2c_adc_address: int = 0x48,
        train_brake_chn: P = None,
        quilling_horn_chn: P = None,
        base_online_pin: P = None,
        base_offline_pin: P = None,
        base_cathode: bool = True,
        base_ping_freq: int = 5,
        lcd_address: int = LCD_PCF8574_ADDRESS,
        lcd_rows: int = 4,
        lcd_cols: int = 20,
        is_lcd: bool = False,
        oled_address: int = 0,
        oled_device="ssd1362",
        is_oled: bool = True,
        keypad: Keypad | KeyPadI2C = None,
        num_lasts: int = 3,
    ):
        self._lock = RLock()
        if is_lcd is True and is_oled is True:
            raise AttributeError("Must specify is_oled=True or is_lcd=True, not both")

        if is_lcd is True and lcd_address:
            self._lcd = Lcd(address=lcd_address, rows=lcd_rows, cols=lcd_cols)
        else:
            self._lcd = None
        if is_oled is True and oled_address is not None:
            self._status = EngineStatus(address=oled_address, device=oled_device)
        else:
            self._status = None

        if row_pins and column_pins:
            self._keypad = Keypad(row_pins, column_pins)
        else:
            self._keypad = keypad
        self._key_queue = self._keypad.key_queue
        self._last_listener = None
        self._state_store = ComponentStateStore.get()
        self._state = None
        self._scope = CommandScope.ENGINE
        self._tmcc_id = None
        self._railroad = None
        self._last_known_speed = None
        self._state_watcher = None
        self._last_motive = UniqueDeque[tuple[int, CommandScope]](maxlen=num_lasts)
        if speed_pins or fwd_pin or rev_pin or reset_pin:
            self._engine_controller = EngineController(
                speed_pin_1=speed_pins[0] if speed_pins and len(speed_pins) > 0 else None,
                speed_pin_2=speed_pins[1] if speed_pins and len(speed_pins) > 1 else None,
                reset_pin=reset_pin,
                halt_pin=halt_pin,
                fwd_pin=fwd_pin,
                rev_pin=rev_pin,
                front_coupler_pin=front_coupler_pin,
                rear_coupler_pin=rear_coupler_pin,
                start_up_pin=start_up_pin,
                shutdown_pin=shutdown_pin,
                boost_pin=boost_pin,
                brake_pin=brake_pin,
                bell_pin=bell_pin,
                horn_pin=horn_pin,
                rpm_up_pin=rpm_up_pin,
                rpm_down_pin=rpm_down_pin,
                labor_up_pin=labor_up_pin,
                labor_down_pin=labor_down_pin,
                vol_up_pin=vol_up_pin,
                vol_down_pin=vol_down_pin,
                smoke_on_pin=smoke_on_pin,
                smoke_off_pin=smoke_off_pin,
                tower_dialog_pin=tower_dialog_pin,
                engr_dialog_pin=engr_dialog_pin,
                aux1_pin=aux1_pin,
                aux2_pin=aux2_pin,
                aux3_pin=aux3_pin,
                stop_immediate_pin=stop_immediate_pin,
                i2c_adc_address=i2c_adc_address,
                train_brake_chn=train_brake_chn,
                quilling_horn_chn=quilling_horn_chn,
                base_online_pin=base_online_pin,
                base_offline_pin=base_offline_pin,
                base_cathode=base_cathode,
                base_ping_freq=base_ping_freq,
            )
        else:
            self._engine_controller = None
        self._is_running = True
        # check for state synchronization
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Controller")
        if self._sync_state and self._sync_state.is_synchronized:
            self._sync_watcher = None
            self.on_sync()
        else:
            self.update_display()
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)

    @property
    def engine_controller(self) -> EngineController:
        return self._engine_controller

    @property
    def is_synchronized(self) -> bool:
        return self._synchronized

    def run(self) -> None:
        while self._is_running:
            key = self._key_queue.wait_for_keypress(60)
            if self._key_queue.is_clear:
                self.process_clear_key()
            elif self._key_queue.is_eol:
                if self._key_queue.key_presses:
                    self.update_engine(self._key_queue.key_presses)
                else:
                    self._key_queue.reset()
            elif key == "A":
                self.change_scope(CommandScope.ENGINE)
            elif key == "B":
                self.change_scope(CommandScope.TRAIN)
            elif key == "*":
                self.last_engine()
            elif key is not None:
                if self._key_queue.is_digit:
                    if self._engine_controller:
                        self._engine_controller.on_numeric(key)
                    self._key_queue.reset()
                else:
                    if self._status:
                        self._status.display.write(str(key))
                        if len(self._key_queue) >= 4:
                            self.update_engine(self._key_queue.key_presses)
                        else:
                            self._status.display.refresh_display()
                    if self._lcd:
                        self._lcd.print(key)
            sleep(0.1)

    def monitor_state_updates(self):
        if self._state_watcher:
            self._state_watcher.shutdown()
            self._state_watcher = None
        if self._state:
            self._state_watcher = StateWatcher(self._state, self.on_state_update)

    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
            self._synchronized = True
            self.update_display()
            self.start()
            self.cache_handler(self)

    def on_state_update(self) -> None:
        cur_speed = self._state.speed if self._state else None
        if cur_speed is not None and self._last_known_speed != cur_speed:
            self._last_known_speed = cur_speed
            if self._engine_controller:
                self._engine_controller.on_speed_changed(cur_speed)
        self.update_display(clear_display=False)

    def cache_engine(self) -> None:
        # don't cache an empty id
        if self._tmcc_id is None or self._state is None or self._tmcc_id != self._state.address:
            return
        # make sure there's a scope
        self._scope = self._scope if self._scope else CommandScope.ENGINE
        # push the most recent engine to the head of the queue (element 0)
        self._last_motive.push((self._tmcc_id, self._scope))

    def last_engine(self):
        if len(self._last_motive):
            # get the item at the head of the list
            if self._tmcc_id is None or self._state is None:
                self._tmcc_id = self._last_motive[0][0]
                self._scope = self._last_motive[0][1]
                self._last_motive.rotate(-1)
            else:
                if self._tmcc_id == self._last_motive[0][0] and self._scope == self._last_motive[0][1]:
                    # if the top of the stack is this engine, try the next one
                    self._last_motive.rotate(-1)
                # else:
                #     # otherwise, save the current engine as the second one in the queue
                #     last_motive = self._last_motive.popleft()
                #     self._last_motive.push((self._tmcc_id, self._scope))
                #     self._last_motive.push(last_motive)
                self._tmcc_id = self._last_motive[0][0]
                self._scope = self._last_motive[0][1]
            self.update_engine(self._tmcc_id)

    def change_scope(self, scope: CommandScope) -> None:
        self.cache_engine()
        self._scope = scope
        self._tmcc_id = self._state = None
        self._key_queue.reset()
        if self._status:
            self._status.update_engine(self._tmcc_id, self._scope)
        self.update_display()

    def process_clear_key(self) -> None:
        self._key_queue.reset()
        if self._tmcc_id and self._state is None:
            # clear not found engine/train
            self._tmcc_id = None
        self.update_engine(self._tmcc_id)
        if self._status:
            self._status.update_engine(self._tmcc_id, self._scope)
        self.update_display()

    def update_engine(self, engine_id: str | int | None):
        if engine_id:
            tmcc_id = int(engine_id)
            # allow use of road numbers; unless an engine supports 4-digit addressing,
            # road numbers are >= 100
            if tmcc_id > 99:
                state = ComponentStateStore.get_state(self._scope, tmcc_id, False)
                if state:
                    tmcc_id = state.address
            # is this engine defined?
            prev_state = self._state
            self._state = self._state_store.get_state(self._scope, tmcc_id, False)
            if self._state:
                if prev_state and self._tmcc_id is not None and tmcc_id != self._tmcc_id:
                    self.cache_engine()
                self._tmcc_id = tmcc_id
                if self._status:
                    self._status.update_engine(self._tmcc_id, self._scope)
                self._last_known_speed = self._state.speed if self._state else None
                if self._engine_controller:
                    self._engine_controller.update(tmcc_id, self._scope, self._state)
            else:
                self._tmcc_id = tmcc_id
            if self._status:
                self._status.update_engine(self._tmcc_id, self._scope)
            self.monitor_state_updates()
            self._key_queue.reset()
            self.cache_engine()
            self.update_display()

    def update_display(self, clear_display: bool = True) -> None:
        if self._lcd is None:
            return
        with self._lock:
            self._lcd.clear_frame_buffer()
            if self._state is not None:
                rname = self._state.road_name if self._state.road_name else "No Information"
                rnum = f"#{self._state.road_number} " if self._state.road_number else ""
                row = f"{rnum}{rname}"
            else:
                row = self.railroad
            self._lcd.add(row)
            if self.is_synchronized:
                row = f"{self._scope.label}: "
                tmcc_id_pos = len(row)
                if self._tmcc_id is not None:
                    row += f"{self._tmcc_id:04}"
                    if self._state:
                        if self._state.control_type is not None:
                            row += f" {self._state.control_type_label}"
                self._lcd.add(row)
                if self._state is not None:
                    if self._lcd.rows > 2:
                        row = f"Speed: {self._state.speed if self._state.speed is not None else 'N/A':03} "
                        row += self._state.direction_label
                        row += f" Ef: {self._state.labor_label:.2} "
                        # row += f"  {self._state.year if self._state.year else ''}"
                        self._lcd.add(row)
                    if self._lcd.rows > 3:
                        row = f"Mom: {self._state.momentum_label:.1} "
                        row += f"TB: {self._state.train_brake_label:.1} "
                        if self._state.is_rpm:
                            row += f"RPM: {self._state.rpm_label:.1}"
                        self._lcd.add(row)
            else:
                tmcc_id_pos = 0

            self._lcd.write_frame_buffer(clear_display)
            if self.is_synchronized and self._tmcc_id is None:
                self._lcd.cursor_pos = (1, tmcc_id_pos)

    @property
    def railroad(self) -> str:
        if self._railroad is None:
            base_state = self._state_store.get_state(CommandScope.BASE, 0, False)
            if base_state and base_state.base_name:
                self._railroad = base_state.base_name.title()
        return self._railroad if self._railroad is not None else "Loading Engine Roster..."

    def reset(self) -> None:
        self._is_running = False
        self._key_queue.reset()
        if self._lcd:
            self._lcd.close(True)
        if self._status:
            self._status.close()
        self._keypad.close()
        if self._state_watcher:
            self._state_watcher.shutdown()

    def close(self) -> None:
        self.reset()


class ControllerI2C(Controller):
    def __init__(
        self,
        keypad_address: int = KEYPAD_PCF8574_ADDRESS,
        speed_pins: List[int | str] = None,
        halt_pin: P = None,
        reset_pin: P = None,
        fwd_pin: P = None,
        rev_pin: P = None,
        front_coupler_pin: P = None,
        rear_coupler_pin: P = None,
        start_up_pin: P = None,
        shutdown_pin: P = None,
        boost_pin: P = None,
        brake_pin: P = None,
        bell_pin: P = None,
        horn_pin: P = None,
        rpm_up_pin: P = None,
        rpm_down_pin: P = None,
        labor_up_pin: P = None,
        labor_down_pin: P = None,
        vol_up_pin: P = None,
        vol_down_pin: P = None,
        smoke_on_pin: P = None,
        smoke_off_pin: P = None,
        tower_dialog_pin: P = None,
        engr_dialog_pin: P = None,
        aux1_pin: P = None,
        aux2_pin: P = None,
        aux3_pin: P = None,
        stop_immediate_pin: P = None,
        i2c_adc_address: int = 0x48,
        train_brake_chn: P = None,
        quilling_horn_chn: P = None,
        base_online_pin: P = None,
        base_offline_pin: P = None,
        base_cathode: bool = True,
        base_ping_freq: int = 5,
        lcd_address: int = 0x27,
        lcd_rows: int = 4,
        lcd_cols: int = 20,
        is_lcd: bool = False,
        oled_address: int = 0,
        oled_device="ssd1362",
        is_oled: bool = True,
        num_lasts: int = 3,
    ):
        keypad = KeyPadI2C(keypad_address)
        super().__init__(
            speed_pins=speed_pins,
            halt_pin=halt_pin,
            reset_pin=reset_pin,
            fwd_pin=fwd_pin,
            rev_pin=rev_pin,
            front_coupler_pin=front_coupler_pin,
            rear_coupler_pin=rear_coupler_pin,
            start_up_pin=start_up_pin,
            shutdown_pin=shutdown_pin,
            boost_pin=boost_pin,
            brake_pin=brake_pin,
            bell_pin=bell_pin,
            horn_pin=horn_pin,
            rpm_up_pin=rpm_up_pin,
            rpm_down_pin=rpm_down_pin,
            labor_up_pin=labor_up_pin,
            labor_down_pin=labor_down_pin,
            vol_up_pin=vol_up_pin,
            vol_down_pin=vol_down_pin,
            smoke_on_pin=smoke_on_pin,
            smoke_off_pin=smoke_off_pin,
            tower_dialog_pin=tower_dialog_pin,
            engr_dialog_pin=engr_dialog_pin,
            aux1_pin=aux1_pin,
            aux2_pin=aux2_pin,
            aux3_pin=aux3_pin,
            stop_immediate_pin=stop_immediate_pin,
            i2c_adc_address=i2c_adc_address,
            train_brake_chn=train_brake_chn,
            quilling_horn_chn=quilling_horn_chn,
            base_online_pin=base_online_pin,
            base_offline_pin=base_offline_pin,
            base_cathode=base_cathode,
            base_ping_freq=base_ping_freq,
            lcd_address=lcd_address,
            lcd_rows=lcd_rows,
            lcd_cols=lcd_cols,
            is_lcd=is_lcd,
            oled_address=oled_address,
            oled_device=oled_device,
            is_oled=is_oled,
            keypad=keypad,
            num_lasts=num_lasts,
        )



from unittest import mock

from src.pytrain.atc.block import Block
from src.pytrain.db.component_state_store import ComponentStateStore
from src.pytrain.protocol.constants import Direction
from test.test_base import TestBase


class TestBlock(TestBase):
    def test_create_block(self) -> None:
        with mock.patch.object(ComponentStateStore, "get_state") as mk_get_state:
            mk_get_state.return_value = None
            with mock.patch.object(Block, "broadcast_state") as mk_broadcast_state:
                b = Block(1, "test block")
                assert mk_get_state.call_count == 2
                mk_broadcast_state.assert_called_once()
                assert b is not None
                assert b.block_id == 1
                assert b.name == "test block"
                assert b.direction == Direction.L2R
                assert b.is_occupied is False
                assert b.occupied_by is None
                assert b.occupied_direction is None
                assert b.sensor_track is None
                assert b.switch is None
                assert b.next_block is None
                assert b.prev_block is None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import Dict, Any

from .component_state import ComponentState, L, P, SCOPE_TO_STATE_MAP
from ..protocol.constants import CommandScope, PROGRAM_NAME
from ..protocol.command_req import CommandReq
from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum


class SyncState(ComponentState):
    """
    Maintain the state of a Lionel Base
    """

    def __init__(self, scope: CommandScope = CommandScope.SYNC) -> None:
        if scope != CommandScope.SYNC:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._state_synchronized: bool | None = None
        self._state_synchronizing: bool | None = None

    def __repr__(self) -> str:
        if self._state_synchronized is not None and not self._state_synchronized:
            msg = "Synchronizing..."
        else:
            msg = f"Synchronized: {self._state_synchronized if self._state_synchronized is not None else 'NA'}"
        return f"{PROGRAM_NAME} {msg}"

    def update(self, command: L | P) -> None:
        if isinstance(command, CommandReq):
            self._ev.clear()
            with self._cv:
                # Note: super().update is explicitly not called
                if command.command in {TMCC1SyncCommandEnum.SYNCHRONIZING, TMCC1SyncCommandEnum.RESYNC}:
                    self._state_synchronized = False
                    self._state_synchronizing = True
                elif command.command == TMCC1SyncCommandEnum.SYNCHRONIZED:
                    self._state_synchronized = True
                    self._state_synchronizing = False
                self.changed.set()
                self._cv.notify_all()

    @property
    def is_synchronized(self) -> bool:
        return self._state_synchronized

    @property
    def is_synchronizing(self) -> bool:
        return self._state_synchronizing

    @property
    def is_known(self) -> bool:
        return self._state_synchronized is not None

    @property
    def is_tmcc(self) -> bool:
        return True

    @property
    def is_legacy(self) -> bool:
        return False

    @property
    def is_lcs(self) -> bool:
        return False

    def as_bytes(self) -> bytes:
        return bytes()

    def as_dict(self) -> Dict[str, Any]:
        state = "synchronized" if self.is_synchronized else "synchronizing" if self.is_synchronizing else None
        return {"state": state}


SCOPE_TO_STATE_MAP.update({CommandScope.SYNC: SyncState})



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

import time
from dataclasses import dataclass
from typing import Any, Callable

import pytest

from src.pytrain.gui import launch_gui as mod
from src.pytrain.protocol.constants import CommandScope

# ---------------------------
# Test doubles (no tkinter)
# Tests generated by ChatGPT 5
# ---------------------------


# noinspection PyMethodMayBeStatic,PyUnusedLocal,PyBroadException
class DummyApp:
    def __init__(self, title: str, width: int, height: int):
        self.title = title
        self.width = width
        self.height = height
        self.bg = "white"
        self._destroyed = False
        self.full_screen = False
        self.when_closed = None

    def after(self, delay_ms: int, func: Callable, args: list | None = None):
        # Run immediately for tests
        if args is None:
            func()
        else:
            func(*args)

    def repeat(self, delay_ms: int, func: Callable):
        # In tests, call once to simulate scheduling
        try:
            func()
        except Exception:
            pass

    def display(self):
        # Return immediately; tests don't want an event loop
        return

    def destroy(self):
        self._destroyed = True


# noinspection PyUnusedLocal
class DummyBox:
    def __init__(self, *args, **kwargs):
        self.enabled = True
        self.visible = True

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def hide(self):
        self.visible = False

    def show(self):
        self.visible = True


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class DummyText:
    def __init__(self, *args, **kwargs):
        self.value = kwargs.get("text", "")
        self.text_color = kwargs.get("color", "black")
        self.visible = True

    def clear(self):
        self.value = ""

    def show(self):
        self.visible = True

    def repeat(self, delay_ms: int, func: Callable):
        # Call once for tests
        func()

    def cancel(self, func: Callable):
        # No-op for tests
        pass


class _DummyTkHandle:
    def config(self, **kwargs):
        # No-op for tests
        pass


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class DummyPushButton:
    def __init__(self, *args, **kwargs):
        self.image = kwargs.get("image", None)
        self.height = kwargs.get("height", 0)
        self.width = kwargs.get("width", 0)
        self.enabled = True
        self.tk = _DummyTkHandle()

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    # noinspection PyMethodMayBeStatic
    def after(self, delay_ms: int, func: Callable, args: list | None = None):
        if args is None:
            func()
        else:
            func(*args)


# CommandReq and dispatcher/store doubles


@dataclass
class SentCall:
    kwargs: dict


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class FakeReq:
    def __init__(self, *args, **kwargs):
        self._sent = []

    def send(self, **kwargs):
        self._sent.append(SentCall(kwargs=kwargs))

    def as_action(self, **_kwargs):
        def act(**_k):
            # record an invocation
            self.send(**_k)

        return act


class FakeDispatcher:
    def __init__(self):
        self.subscribed = []

    def subscribe(self, who: Any, scope, addr: int):
        self.subscribed.append((who, scope, addr))


class FakeSyncState:
    def __init__(self, synchronized: bool = True):
        self._sync = synchronized

    @property
    def is_synchronized(self) -> bool:
        return self._sync


class FakeEngineState:
    def __init__(self, started=False, aux2=False):
        self.is_started = started
        self.is_aux2 = aux2


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class FakeStore:
    def __init__(self, sync=True, started=False, aux2=False):
        self._sync = FakeSyncState(sync)
        self._engine = FakeEngineState(started, aux2)

    def get_state(self, scope, addr, *args, **kwargs):
        if scope == CommandScope.SYNC:
            return self._sync
        if scope == CommandScope.ENGINE:
            return self._engine
        return None


class DummyWatcher:
    def __init__(self, *_args, **_kwargs):
        pass

    def shutdown(self):
        pass


# ---------------------------
# Fixtures
# ---------------------------


@pytest.fixture(autouse=True)
def patch_guizero_and_system(monkeypatch):
    # Replace guizero widgets with mock objects
    monkeypatch.setattr(mod, "App", DummyApp, raising=True)
    monkeypatch.setattr(mod, "Box", DummyBox, raising=True)
    monkeypatch.setattr(mod, "Text", DummyText, raising=True)
    monkeypatch.setattr(mod, "PushButton", DummyPushButton, raising=True)

    # Replace CommandReq with our recorder
    monkeypatch.setattr(mod, "CommandReq", FakeReq, raising=True)

    # Replace dispatcher and store singletons
    fake_dispatcher = FakeDispatcher()
    monkeypatch.setattr(mod.CommandDispatcher, "get", staticmethod(lambda: fake_dispatcher), raising=True)

    fake_store = FakeStore(sync=True, started=False, aux2=False)
    monkeypatch.setattr(mod.ComponentStateStore, "get", staticmethod(lambda: fake_store), raising=True)
    monkeypatch.setattr(mod.ComponentStateStore, "get_state", staticmethod(fake_store.get_state), raising=True)

    # Replace StateWatcher with no-op
    monkeypatch.setattr(mod, "StateWatcher", DummyWatcher, raising=True)

    # Replace GpioHandler cache to avoid side effects
    monkeypatch.setattr(mod.GpioHandler, "cache_handler", staticmethod(lambda *_: None), raising=True)

    # Replace find_file so we don't need assets
    monkeypatch.setattr(mod, "find_file", lambda name: name, raising=True)

    return fake_dispatcher, fake_store


@pytest.fixture
def instance(monkeypatch):
    # Prevent thread.run from building full UI for some tests by patching run to a no-op.
    # Individual tests can override if they want the UI structure.
    monkeypatch.setattr(mod.LaunchGui, "run", lambda self: None, raising=True)
    return mod.LaunchGui(tmcc_id=39, width=640, height=480)


# ---------------------------
# Tests
# ---------------------------


def test_scale_calculation(instance):
    # width set to 640; scale should proportionally increase from base 480
    # value=72: 72 * 640/480 = 96; with factor=0.7 -> max(72, int(0.7*96)) = max(72, 67) = 72 (since >480 uses factor)
    # but width>480 so factor applies on the already scaled result; 96*0.7 = 67 -> min bound is orig (72)
    assert instance.scale(72, 0.7) == 72
    # For bigger base ensure it grows
    assert instance.scale(128) == max(128, int(128 * 640 / 480))  # 170


def test_on_sync_starts_and_subscribes(monkeypatch):
    started = {"flag": False}

    # noinspection PyUnusedLocal
    def fake_run(self):
        started["flag"] = True

    monkeypatch.setattr(mod.LaunchGui, "run", fake_run, raising=True)

    inst = mod.LaunchGui(tmcc_id=41, width=320, height=240)

    # on_sync should be triggered (sync True), setting started flag
    assert started["flag"] is True

    # And subscribe should be called
    # Fetch the fake dispatcher installed in fixture
    dispatcher = mod.CommandDispatcher.get()
    assert (inst, CommandScope.ENGINE, 41) in dispatcher.subscribed


def test_run_builds_min_ui_and_sync_state(monkeypatch):
    # Let the real run execute against mock objects to create widget tree
    inst = mod.LaunchGui(tmcc_id=39, width=640, height=480)
    # Give some time for thread to run setup and immediately return
    time.sleep(0.05)

    # After run, UI is torn down in finally, so app is None
    assert inst.app is None
    # But run should have executed without exceptions using mock classes


def test_toggle_power_on_path(monkeypatch):
    # Use a trimmed-down run to install minimal widget graph and not tear it down
    def mini_run(self):
        self.app = DummyApp("Launch Pad", self.width, self.height)
        self.lower_box = DummyBox()
        self.upper_box = DummyBox()
        self.lights_box = DummyBox()
        self.siren_box = DummyBox()
        self.klaxon_box = DummyBox()
        self.gantry_box = DummyBox()
        self.comms_box = DummyBox()
        self.power_button = DummyPushButton(image=self.on_button, height=self.s_72, width=self.s_72)
        self.lights_button = DummyPushButton(image=self.on_button, height=self.s_72, width=self.s_72)
        self.label = DummyText(text="T-Minus")
        self.message = DummyText(text="")
        self.count = DummyText(text="")
        self.launch = DummyPushButton()
        # monitored state used by sync_pad_lights
        self._monitored_state = FakeEngineState(started=True, aux2=False)

    monkeypatch.setattr(mod.LaunchGui, "run", mini_run, raising=True)

    inst = mod.LaunchGui(tmcc_id=55, track_id=10, width=640, height=480)

    # Initial state: power button shows "on" image
    assert inst.power_button.image == inst.on_button

    inst.toggle_power()

    # After toggling on: image should switch to off icon
    assert inst.power_button.image == inst.off_button
    # lower_box visibility cycled back on
    assert inst.lower_box.visible is True

    # Power-on command should have been sent with delay=0.5
    assert len(inst.power_on_req._sent) == 1
    assert pytest.approx(inst.power_on_req._sent[0].kwargs.get("delay", 0.0), 0.01) == 0.5

    # Track on request also sent (because track_id provided)
    assert inst.track_on_req is not None
    assert len(inst.track_on_req._sent) == 1


def test_toggle_power_off_path(monkeypatch):
    def mini_run(self):
        self.app = DummyApp("Launch Pad", self.width, self.height)
        self.lower_box = DummyBox()
        self.upper_box = DummyBox()
        self.lights_box = DummyBox()
        self.siren_box = DummyBox()
        self.klaxon_box = DummyBox()
        self.gantry_box = DummyBox()
        self.comms_box = None
        self.power_button = DummyPushButton(image=self.off_button, height=self.s_72, width=self.s_72)
        self.lights_button = DummyPushButton(image=self.on_button, height=self.s_72, width=self.s_72)
        self.klaxon_button = DummyPushButton(image=self.siren_on, height=self.s_72, width=self.s_72)
        self.label = DummyText(text="T-Minus")
        self.message = DummyText(text="")
        self.count = DummyText(text="")
        self.launch = DummyPushButton()
        self._monitored_state = FakeEngineState(started=False, aux2=True)

    monkeypatch.setattr(mod.LaunchGui, "run", mini_run, raising=True)

    inst = mod.LaunchGui(tmcc_id=60, width=640, height=480)

    # Now toggling should power off
    inst.toggle_power()

    assert inst.power_button.image == inst.on_button
    assert len(inst.power_off_req._sent) == 1


def test_set_lights_icons_and_klaxon_off(monkeypatch):
    def mini_run(self):
        self.app = DummyApp("Launch Pad", self.width, self.height)
        self.lower_box = DummyBox()
        self.power_button = DummyPushButton(image=self.on_button, height=self.s_72, width=self.s_72)
        self.lights_button = DummyPushButton(image=self.on_button, height=self.s_72, width=self.s_72)
        self.klaxon_button = DummyPushButton(image=self.siren_on, height=self.s_72, width=self.s_72)

    monkeypatch.setattr(mod.LaunchGui, "run", mini_run, raising=True)

    inst = mod.LaunchGui(tmcc_id=61, width=640, height=480)

    # Turn lights off icon
    inst.set_lights_off_icon()
    assert inst.lights_button.image == inst.off_button
    assert inst.lights_button.height == inst.s_72
    assert inst.lights_button.width == inst.s_72

    # Turn lights on icon
    inst.set_lights_on_icon()
    assert inst.lights_button.image == inst.on_button

    # Klaxon off updates image/size
    inst.set_klaxon_off_icon()
    assert inst.klaxon_button.image == inst.siren_off
    assert inst.klaxon_button.height == inst.s_72
    assert inst.klaxon_button.width == inst.s_72



import logging
import math
import sched
import threading
import time
from threading import Thread
from typing import Callable, Dict, Tuple, TypeVar, Union, cast

from gpiozero import LED, MCP3008, MCP3208, AnalogInputDevice, Button, Device

from ..comm.comm_buffer import CommBuffer
from ..comm.command_listener import Message
from ..db.component_state_store import DependencyCache
from ..db.engine_state import EngineState
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import DEFAULT_ADDRESS, PROGRAM_NAME, CommandScope
from ..protocol.tmcc1.tmcc1_constants import (
    TMCC1AuxCommandEnum,
)
from .i2c.ads_1x15 import Ads1115
from .i2c.button_i2c import ButtonI2C
from .i2c.led_i2c import LEDI2C

log = logging.getLogger(__name__)

DEFAULT_BOUNCE_TIME: float = 0.015  # button debounce threshold
DEFAULT_VARIANCE: float = 0.001  # pot difference variance

T = TypeVar("T", bound=CommandReq)

P = TypeVar("P", bound=Union[int, str, Tuple[int], Tuple[int, int], Tuple[int, int, int]])


class PressedHeldDef:
    def __init__(
        self,
        pressed_req: CommandReq,
        held_req: CommandReq = None,
        repeat_action: int = 1,
        held_threshold: float = 0.5,
        hold_repeat: bool = False,
        frequency: float = 0.1,
        data: Callable = None,
    ) -> None:
        self._pressed_req = pressed_req
        self._held_req = held_req if held_req else pressed_req
        self._repeat_action = repeat_action
        self.held_threshold = held_threshold
        self.repeat = hold_repeat
        self.frequency = frequency
        self._data_gen = data

    def update_target(self, address: int = None, data: int = None, scope: CommandScope = None) -> None:
        if address is not None:
            self._pressed_req.address = address
            if self._pressed_req != self._held_req:
                self._held_req.address = address
        if data is not None:
            self._pressed_req.data = data
            if self._pressed_req != self._held_req:
                self._held_req.data = data
        if scope is not None:
            self._pressed_req.scope = scope
            if self._pressed_req != self._held_req:
                self._held_req.scope = scope

    def as_action(self, address: int = None, data: int = None, scope: CommandScope = None):
        self.update_target(address=address, data=data, scope=scope)
        return GpioHandler.when_button_pressed_or_held_action(
            self._pressed_req.as_action(repeat=self._repeat_action),
            self._held_req.as_action(repeat=self._repeat_action),
            held_threshold=self.held_threshold,
            held_repeat=self.repeat,
            frequency=self.frequency,
            data_gen=self._data_gen,
        )


class GpioDelayHandler(Thread):
    """
    Handle delayed (scheduled) requests. Implementation uses Python's lightweight
    sched module to keep a list of requests to issue in the future. We use
    threading.Event.wait() as the sleep function, as it is interruptable. This
    allows us to schedule requests in any order and still have them fire at the
    appropriate time.
    """

    def __init__(self) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} GPIO Delay Handler")
        self._cv = threading.Condition()
        self._ev = threading.Event()
        self._scheduler = sched.scheduler(time.time, self._ev.wait)
        self._running = True
        self.start()

    def cancel(self, ev: sched.Event) -> None:
        try:
            if ev:
                self._scheduler.cancel(ev)
        except ValueError:
            pass

    def reset(self) -> None:
        self._running = False
        # clear out any scheduled events
        for ev in self._scheduler.queue:
            self.cancel(ev)
        # send shutdown signal
        with self._cv:
            self._ev.set()
            self._cv.notify()

    def run(self) -> None:
        while self._running:
            with self._cv:
                while self._running and self._scheduler.empty():
                    self._cv.wait()
            # run the scheduler outside the cv lock; otherwise,
            # we couldn't schedule more commands
            self._scheduler.run()
            self._ev.clear()

    def schedule(self, delay: float, command: Callable) -> sched.Event:
        with self._cv:
            ev = self._scheduler.enter(delay, 1, command)
            # this interrupts the running scheduler
            self._ev.set()
            # and this notifies the main thread to restart, as there is a new
            # request in the sched queue
            self._cv.notify()
            return ev


class PotHandler(Thread):
    def __init__(
        self,
        command: CommandReq | None,
        channel: int = 0,
        use_12bit: bool = False,
        data_min: int = None,
        data_max: int = None,
        threshold: float = None,
        delay: float = 0.05,
        scale: Dict[int, int] = None,
        cmds: Dict[int, T] = None,
        start: bool = True,
        prefix: CommandReq = None,
    ) -> None:
        super().__init__(daemon=True)
        if use_12bit:
            self._pot = MCP3208(channel=channel, differential=False)
        else:
            self._pot = MCP3008(channel=channel)
        self._command = command
        self._prefix = prefix
        self._prefix_data = prefix.data if prefix else None
        self._prefix_address = prefix.address if prefix else None
        self._last_value = None
        self._action = command.as_action() if command else None
        self._data_max = data_max = data_max if data_max is not None else command.data_max
        self._data_min = data_min = data_min if data_min is not None else command.data_min
        self._interp = GpioHandler.make_interpolator(data_max, data_min)
        if threshold is not None:
            self._threshold = threshold
        elif command:
            self._threshold = 1 if command.num_data_bits < 6 else 2
        else:
            self._threshold = None
        self._delay = delay
        self._running = True
        self._scale = scale
        self._command_map = cmds
        self._tmcc_command_buffer = CommBuffer.build()
        if start:
            self.start()

    @property
    def pot(self) -> AnalogInputDevice:
        return self._pot

    def run(self) -> None:
        if log.isEnabledFor(logging.DEBUG):
            log.debug(
                f"Delay: {self._delay} threshold: {self._threshold} d_min: {self._data_min} d_max: {self._data_max}"
            )
        while self._running:
            raw_value = self.pot.value
            value = self._interp(raw_value)
            if self._scale:
                value = self._scale[value]
            if self._last_value is None:
                self._last_value = value
                continue
            elif self._threshold is not None and math.fabs(self._last_value - value) <= self._threshold and value > 0:
                continue  # pots can take a bit to settle; ignore small changes
            if self._last_value == 0 and value == 0:
                continue
            self._last_value = value
            byte_str = bytes()
            if self._command_map and value in self._command_map:
                cmd = self._command_map[value]
                if cmd:
                    if self._prefix:
                        if GpioHandler.engine_numeric(self._prefix_address) == self._prefix_data:
                            pass
                        else:
                            byte_str += self._prefix.as_bytes * 3
                    # command could be None, indicating no action
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug(f"{cmd} {value} {raw_value}")
                    if cmd.is_data:
                        cmd.data = value
                    byte_str += cmd.as_bytes
            elif self._command and self._action:
                if self._prefix:
                    if GpioHandler.engine_numeric(self._prefix_address) == self._prefix_data:
                        pass
                    else:
                        byte_str += self._prefix.as_bytes * 3
                self._command.data = value
                byte_str += self._command.as_bytes
            if byte_str:
                self._tmcc_command_buffer.enqueue_command(byte_str)
            time.sleep(self._delay)

    def reset(self) -> None:
        self._running = False


class JoyStickHandler(PotHandler):
    def __init__(
        self,
        command: CommandReq | None = None,
        channel: int = 0,
        use_12bit: bool = True,
        data_min: int = None,
        data_max: int = None,
        delay: float = 0.05,
        scale: Dict[int, int] = None,
        cmds: Dict[int, T] = None,
        prefix: CommandReq = None,
    ) -> None:
        super().__init__(
            command,
            channel=channel,
            use_12bit=use_12bit,
            data_min=data_min,
            data_max=data_max,
            threshold=None,
            delay=delay,
            scale=scale,
            cmds=cmds,
            start=False,
            prefix=prefix,
        )
        self._threshold = None
        self.start()


# noinspection DuplicatedCode
class GpioHandler:
    GPIO_DEVICE_CACHE = set()
    GPIO_HANDLER_CACHE = set()
    GPIO_DELAY_HANDLER = None

    @staticmethod
    def current_milli_time() -> int:
        """
        Return the current time, in milliseconds past the "epoch"
        """
        return round(time.time() * 1000)

    @staticmethod
    def engine_numeric(address: int) -> int | None:
        if address is not None:
            from ..db.component_state_store import ComponentStateStore

            state = ComponentStateStore.get_state(CommandScope.ENGINE, address, create=False)
            if state:
                return cast(EngineState, state).numeric
        return None

    @staticmethod
    def make_interpolator(
        to_max: int | float,
        to_min: int | float = 0,
        from_min: int | float = 0.0,
        from_max: int | float = 1.0,
        as_float: bool = False,
    ) -> Callable:
        # Figure out how 'wide' each range is
        from_span = from_max - from_min
        to_span = to_max - to_min

        # Compute the scale factor between left and right values
        scale_factor = float(to_span) / float(from_span)

        # create interpolation function using pre-calculated scaleFactor
        def interp_fn(value) -> int | float:
            if as_float:
                scaled_value = float((to_min + (value - from_min) * scale_factor))
            else:
                scaled_value = int(round(to_min + (value - from_min) * scale_factor))
            scaled_value = min(scaled_value, to_max)
            scaled_value = max(scaled_value, to_min)
            return scaled_value

        return interp_fn

    @classmethod
    def calibrate_joystick(
        cls,
        x_axis_chn: int = 0,
        y_axis_chn: int = 1,
        i2c_addr=None,
        use_12bit: bool = True,
    ) -> None:
        if i2c_addr:
            x_axis = Ads1115(channel=x_axis_chn, address=i2c_addr)
            y_axis = Ads1115(channel=y_axis_chn, address=i2c_addr)
        elif use_12bit:
            x_axis = MCP3208(channel=x_axis_chn, differential=False)
            y_axis = MCP3208(channel=y_axis_chn, differential=False)
        else:
            x_axis = MCP3008(channel=x_axis_chn, differential=False)
            y_axis = MCP3008(channel=y_axis_chn, differential=False)

        min_x = min_y = float("inf")
        max_x = max_y = float("-inf")
        print("Rotate Joystick clockwise, making sure to fully exercise its range...")
        time.sleep(5)
        is_running = True
        start_at = cls.current_milli_time()
        cycle = 0
        while is_running:
            x = x_axis.value
            y = y_axis.value

            if x < min_x:
                min_x = x
            if x > max_x:
                max_x = x
            if y < min_y:
                min_y = y
            if y > max_y:
                max_y = y
            cycle += 1

            elapsed = cls.current_milli_time() - start_at
            if cycle % 1000 == 0:
                remaining = (10000 - elapsed) / 1000.0
                print(f"Time remaining: {remaining} seconds", end="\r")
            if elapsed > 10000:
                is_running = False
        print(f" X axis range: {min_x} - {max_x}")
        print(f" Y axis range: {min_y} - {max_y}")

        print("Now take your hands off hit joystick and let it recenter for 10 seconds")
        time.sleep(5)
        is_running = True
        min_x = min_y = float("inf")
        max_x = max_y = float("-inf")
        start_at = cls.current_milli_time()
        cycle = 0
        sum_x = sum_y = 0.0
        while is_running:
            x = x_axis.value
            y = y_axis.value

            if x < min_x:
                min_x = x
            if x > max_x:
                max_x = x
            sum_x += x
            if y < min_y:
                min_y = y
            if y > max_y:
                max_y = y
            sum_y += y
            cycle += 1

            elapsed = cls.current_milli_time() - start_at
            if cycle % 1000 == 0:
                remaining = (10000 - elapsed) / 1000.0
                print(f"Time remaining: {remaining} seconds", end="\r")
            if elapsed >= 10000:
                is_running = False
        print(f" X center range: {min_x} - {max_x}; average: {sum_x / cycle}")
        print(f" Y center range: {min_y} - {max_y}; average: {sum_y / cycle}")

    @classmethod
    def accessory(
        cls,
        address: int,
        aux1_pin: P,
        aux2_pin: P,
        aux1_led_pin: P = None,
        cathode: bool = True,
    ) -> Tuple[Button, Button] | Tuple[Button, Button, LED]:
        """
        Control an accessory that responds to TMCC1 accessory commands, such as one connected
        to an LCS ASC2 configured in accessory, 8 TMCC IDs mode.

        Press and hold the Aux1 button to operate the accessory for as long as Aux1 is held.
        Press the Aux2 button to turn the accessory on or off.
        """
        # make the CommandReqs
        aux1_req, aux1_btn, aux1_led = cls.make_button(
            aux1_pin,
            TMCC1AuxCommandEnum.AUX1_OPT_ONE,
            address,
            led_pin=aux1_led_pin,
            cathode=cathode,
            bind=True,
        )
        aux2_req, aux2_btn, aux2_led = cls.make_button(
            aux2_pin,
            TMCC1AuxCommandEnum.AUX2_OPT_ONE,
            address,
            cathode=cathode,
        )
        # bind actions to buttons
        aux1_action = aux1_req.as_action()
        aux2_action = aux2_req.as_action()

        aux1_btn.when_pressed = cls.with_held_action(aux1_action, aux1_btn)
        aux2_btn.when_pressed = aux2_action

        if aux2_led is None:
            return aux1_btn, aux2_btn
        else:
            return aux1_btn, aux2_btn, aux1_led

    @classmethod
    def make_button(
        cls,
        pin: P,
        command: CommandReq | CommandDefEnum = None,
        address: int = DEFAULT_ADDRESS,
        data: int = None,
        scope: CommandScope = None,
        led_pin: P = None,
        hold_repeat: bool = False,
        hold_time: float = 1,
        initially_on: bool = False,
        bind: bool = False,
        cathode: bool = True,
    ) -> Button | Tuple[CommandReq, Button, LED]:
        # if command is actually a CommandDefEnum, build_req a CommandReq
        if isinstance(command, CommandDefEnum):
            command = CommandReq.build(command, address=address, data=data, scope=scope)

        # create the button object we will associate an action with
        hold_time = hold_time if hold_time is not None and hold_time > 0 else 1.0
        if isinstance(pin, tuple):
            # noinspection PyTypeChecker
            button = ButtonI2C(pin, bounce_time=DEFAULT_BOUNCE_TIME, hold_repeat=hold_repeat, hold_time=hold_time)
        else:
            button = Button(pin, bounce_time=DEFAULT_BOUNCE_TIME, hold_repeat=hold_repeat, hold_time=hold_time)
        cls.cache_device(button)

        # create a LED, if asked, and tie its source to the button
        if led_pin is not None:
            led = cls.make_led(led_pin, initially_on, cathode)
            cls.cache_device(led)
            if bind:
                led.source = button
        else:
            led = None

        if command is None:
            return button
        else:
            return command, button, led

    @classmethod
    def make_led(
        cls,
        pin: P,
        initially_on: bool = False,
        cathode: bool = True,
    ) -> LED:
        if isinstance(pin, tuple):
            # noinspection PyTypeChecker
            led = LEDI2C(pin, cathode=cathode, initial_value=initially_on)
        else:
            led = LED(pin, active_high=cathode, initial_value=initially_on)
        return led

    @classmethod
    def when_button_pressed(
        cls,
        pin: P,
        command: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        led_pin: P = None,
        cathode: bool = True,
    ) -> Tuple[CommandReq, Button, LED]:
        # Use helper method to construct objects
        command, button, led = cls.make_button(
            pin,
            command,
            address,
            data,
            scope,
            led_pin,
            cathode=cathode,
        )

        # create a command function to fire when button pressed
        button.when_pressed = command.as_action()
        return command, button, led

    @classmethod
    def when_button_held(
        cls,
        pin: P,
        command: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        frequency: float = 1,
        led_pin: P = None,
        cathode: bool = True,
    ) -> Button:
        # Use helper method to construct objects
        command, button, led = cls.make_button(
            pin,
            command,
            address,
            data=data,
            scope=scope,
            led_pin=led_pin,
            cathode=cathode,
        )

        # create a command function to fire when button held
        button.when_held = command.as_action()
        button.hold_repeat = True
        button.hold_time = frequency
        return button

    @classmethod
    def when_button_pressed_or_held_action(
        cls,
        pressed_action,
        held_action,
        held_threshold,
        held_repeat: bool = False,
        frequency: float = 0.1,
        data_gen: Callable = None,
    ) -> Callable:
        def func(btn: Button) -> None:
            # sleep for hold threshold, if button still active, do held_action
            # otherwise do pressed_action
            trigger_effects = True
            time.sleep(held_threshold)
            if btn.is_active is True:
                while btn.is_active:
                    if data_gen is not None:
                        held_action(new_data=data_gen(btn), trigger_effects=trigger_effects)
                    else:
                        held_action(trigger_effects=trigger_effects)
                    # if held_repeat is true, continue sending when_held action
                    # with the given frequency of repeat
                    if held_repeat:
                        time.sleep(frequency)
                        trigger_effects = False
                    else:
                        break
            else:
                pressed_action()

        return func

    @classmethod
    def when_toggle_switch(
        cls,
        off_pin: P,
        on_pin: P,
        off_command: CommandReq,
        on_command: CommandReq,
        led_pin: P = None,
        cathode: bool = True,
    ) -> Tuple[Button, Button, LED]:
        # Create a LED, if requested. It is turned on by pressing the
        # ON button, and turned off by pressing the OFF button
        if led_pin is not None and led_pin != 0:
            # noinspection PyTypeChecker
            led = cls.make_led(led_pin, cathode=cathode)
            led.on()
        else:
            led = None

        # create the off and on buttons
        off_button = Button(off_pin, bounce_time=DEFAULT_BOUNCE_TIME)
        on_button = Button(on_pin, bounce_time=DEFAULT_BOUNCE_TIME)

        # bind them to functions; we need to wrap the functions if we're using a LED
        off_action = off_command.as_action()
        on_action = on_command.as_action()
        if led is not None:
            off_button.when_pressed = cls.with_off_action(off_action, led)
            on_button.when_pressed = cls.with_on_action(on_action, led)

            def func_off(_: Message) -> None:
                led.off()

            DependencyCache.listen_for_disablers(on_command, func_off)

            def func_on(_: Message) -> None:
                led.on()

            DependencyCache.listen_for_enablers(on_command, func_on)

        else:
            off_button.when_pressed = off_action
            on_button.when_pressed = on_action

        cls.cache_device(off_button)
        cls.cache_device(on_button)
        if led is not None:
            cls.cache_device(led)
        return off_button, on_button, led

    @classmethod
    def when_toggle_button_pressed(
        cls,
        pin: P,
        command: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        led_pin: P = None,
        initial_state: bool = False,
        auto_timeout: int = None,
        cathode: bool = True,
    ) -> tuple[Button, LED]:
        # Use helper method to construct objects
        command, button, led = cls.make_button(
            pin,
            command,
            address,
            data,
            scope,
            led_pin,
            cathode=cathode,
        )

        # create a command function to fire when button pressed
        action = command.as_action()
        if led_pin is not None and led_pin != 0:
            button.when_pressed = cls.with_toggle_action(action, led, auto_timeout)
            led.source = None  # want led to stay lit when button pressed
            if initial_state:
                led.on()
            else:
                led.off()
        else:
            button.when_pressed = action
        return button, led

    @classmethod
    def when_pot(
        cls,
        command: CommandReq | CommandDefEnum | None = None,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = None,
        channel: int = 0,
        use_12bit: bool = True,
        data_min: int = None,
        data_max: int = None,
        threshold: float = None,
        delay: float = 0.05,
        scale: Dict[int, int] = None,
        cmds: Dict[int, T] = None,
    ) -> PotHandler:
        if isinstance(command, CommandDefEnum):
            command = CommandReq.build(command, address, 0, scope)
        if command and command.num_data_bits == 0:
            raise ValueError("Command does not support variable data")
        knob = PotHandler(
            command=command,
            channel=channel,
            use_12bit=use_12bit,
            data_min=data_min,
            data_max=data_max,
            threshold=threshold,
            delay=delay,
            scale=scale,
            cmds=cmds,
        )
        cls.cache_handler(knob)
        cls.cache_device(knob.pot)
        return knob

    @classmethod
    def when_joystick(
        cls,
        command: CommandReq | CommandDefEnum | None = None,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = None,
        channel: int = 0,
        use_12bit: bool = False,
        data_min: int = None,
        data_max: int = None,
        delay: float = 0.05,
        scale: Dict[int, int] = None,
        cmds: Dict[int, T] = None,
        prefix: CommandReq = None,
    ) -> JoyStickHandler:
        if isinstance(command, CommandDefEnum):
            command = CommandReq.build(command, address, 0, scope)
        if command and command.num_data_bits == 0:
            raise ValueError("Command does not support variable data")
        joystick = JoyStickHandler(
            command=command,
            channel=channel,
            use_12bit=use_12bit,
            data_min=data_min,
            data_max=data_max,
            delay=delay,
            scale=scale,
            cmds=cmds,
            prefix=prefix,
        )
        cls.cache_handler(joystick)
        cls.cache_device(joystick.pot)
        return joystick

    @classmethod
    def reset_all(cls) -> None:
        for handler in cls.GPIO_HANDLER_CACHE:
            if not hasattr(handler, "reset"):
                log.error(f"{handler} has no 'reset' method. Skipping...")
                continue
            handler.reset()
            if isinstance(handler, Thread) and handler.is_alive():
                handler.join()  # wait for thread to shut down
        cls.GPIO_HANDLER_CACHE = set()

        for device in cls.GPIO_DEVICE_CACHE:
            device.close()
        cls.GPIO_DEVICE_CACHE = set()

    @classmethod
    def cache_handler(cls, handler: Thread) -> None:
        cls.GPIO_HANDLER_CACHE.add(handler)

    @classmethod
    def release_handler(cls, handler: Thread) -> None:
        if handler in cls.GPIO_HANDLER_CACHE:
            if not hasattr(handler, "reset"):
                log.error(f"{handler} has no 'reset' method. Skipping...")
            else:
                # noinspection PyUnresolvedReferences
                handler.reset()
            cls.GPIO_HANDLER_CACHE.discard(handler)

    @classmethod
    def cache_device(cls, device: Device) -> None:
        """
        Keep devices around after creation so they remain in scope
        """
        cls.GPIO_DEVICE_CACHE.add(device)

    @classmethod
    def release_device(cls, device: Device) -> None:
        device.close()
        cls.GPIO_DEVICE_CACHE.discard(device)

    @classmethod
    def with_held_action(cls, action: Callable, button: Button, delay: float = 0.10) -> Callable:
        def held_action() -> None:
            while button.is_active:
                action()
                time.sleep(delay)

        return held_action

    @classmethod
    def with_toggle_action(cls, action: Callable, led: LED, auto_timeout: int = None) -> Callable:
        if cls.GPIO_DELAY_HANDLER is None:
            cls.GPIO_DELAY_HANDLER = GpioDelayHandler()
            cls.cache_handler(cls.GPIO_DELAY_HANDLER)

        ev: sched.Event | None = None

        def toggle_action() -> None:
            nonlocal ev
            action()
            if led.value:
                if ev is not None:
                    cls.GPIO_DELAY_HANDLER.cancel(ev)
                    ev = None
                led.off()
            else:
                led.on()
                if auto_timeout:
                    ev = cls.GPIO_DELAY_HANDLER.schedule(auto_timeout, led.off)

        return toggle_action

    @classmethod
    def with_off_action(cls, action: Callable, led: LED = None, *impacted_leds: LED) -> Callable:
        def off_action() -> None:
            action()
            if led is not None:
                led.off()
            if impacted_leds:
                for impacted_led in impacted_leds:
                    impacted_led.on()

        return off_action

    @classmethod
    def with_on_action(cls, action: Callable, led: LED, *impacted_leds: LED) -> Callable:
        def on_action() -> None:
            action()
            if led is not None:
                led.on()
            if impacted_leds:
                for impacted_led in impacted_leds:
                    if impacted_led is not None:
                        impacted_led.off()

        return on_action

    @classmethod
    def with_prefix_action(
        cls,
        prefix: CommandReq,
        command: CommandReq,
    ) -> Callable:
        def func() -> None:
            if prefix:
                prefix.send(repeat=2)
            command.send()

        return func



from __future__ import annotations

import abc
import ipaddress
import logging
import sched
import socket
import sys
import threading
import time
import uuid
from ipaddress import IPv4Address, IPv6Address
from queue import Empty, Queue
from threading import Condition, Event, Lock, Thread

from ..db.component_state import ComponentState
from ..pdi.pdi_req import PdiReq
from ..protocol.command_req import CommandReq
from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum

if sys.version_info >= (3, 11):
    from typing import Dict, Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

import serial
from serial.serialutil import SerialException

from ..protocol.constants import (
    DEFAULT_BAUDRATE,
    DEFAULT_PORT,
    DEFAULT_PULSE,
    DEFAULT_QUEUE_SIZE,
    DEFAULT_SER2_THROTTLE_DELAY,
    DEFAULT_SERVER_PORT,
    DEFAULT_VALID_BAUDRATES,
    PROGRAM_NAME,
)

log = logging.getLogger(__name__)


class CommBuffer(abc.ABC):
    from ..db.component_state import ComponentState

    __metaclass__ = abc.ABCMeta

    _instance = None
    _lock = Lock()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if CommBuffer._instance is None:
                CommBuffer._instance = super(CommBuffer, cls).__new__(cls)
                CommBuffer._instance._initialized = False
            return CommBuffer._instance

    @staticmethod
    def parse_server(
        server: str | IPv4Address | IPv6Address, port: str | int, server_port: int = 0
    ) -> tuple[IPv4Address | IPv6Address | None, str]:
        if server is not None:
            try:
                if isinstance(server, str):
                    server = ipaddress.ip_address(socket.gethostbyname(server))
                if server_port > 0:
                    port = server_port
                elif isinstance(port, str) and not port.isnumeric():
                    port = str(DEFAULT_SERVER_PORT)
            except Exception as e:
                log.error(f"Failed to resolve {server}: {e}")
                log.exception(e)
                raise e
        return server, port

    @classmethod
    def build(
        cls,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
        ser2=False,
    ) -> Self:
        if cls._instance:
            return cls._instance
        """
        We only want one or the other of these buffers per process
        """
        server, port = cls.parse_server(server, port)
        if server is None:
            return CommBufferSingleton(queue_size=queue_size, baudrate=baudrate, port=port, ser2=ser2)
        else:
            return CommBufferProxy(server, int(port))

    @classmethod
    def get(cls) -> CommBuffer:
        if cls._instance is None:
            raise AttributeError("CommBuffer has not been initialized")
        return cls._instance

    @classmethod
    def stop(cls) -> None:
        with cls._lock:
            if cls._instance is not None:
                cls._instance.shutdown()

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def is_server(cls) -> bool:
        if not cls.is_built():
            raise ValueError("CommBuffer is not built")
        return isinstance(cls._instance, CommBufferSingleton)

    @classmethod
    def is_client(cls) -> bool:
        return isinstance(cls._instance, CommBufferProxy)

    @classmethod
    def no_ser2(cls) -> bool:
        return True

    @classmethod
    def server_port(cls) -> int | None:
        return None

    @classmethod
    def server_ip(cls) -> str | None:
        return None

    @abc.abstractmethod
    def enqueue_command(self, command: bytes, delay: float = 0) -> None:
        """
        Enqueue the command to send to the Lionel LCS SER2
        """
        ...

    @abc.abstractmethod
    def update_state(self, state: ComponentState | CommandReq | PdiReq | bytes) -> None:
        """
        Update all nodes with the state change
        """
        ...

    @abc.abstractmethod
    def shutdown(self, immediate: bool = False) -> None: ...

    @abc.abstractmethod
    def register(self, port: int = DEFAULT_SERVER_PORT) -> None: ...

    @abc.abstractmethod
    def disconnect(self, port: int = DEFAULT_SERVER_PORT) -> None: ...

    @abc.abstractmethod
    def sync_state(self, port: int = DEFAULT_SERVER_PORT) -> None: ...

    @abc.abstractmethod
    def start_heart_beat(self, port: int = DEFAULT_SERVER_PORT): ...

    @abc.abstractmethod
    def join(self) -> None: ...

    @property
    @abc.abstractmethod
    def server_version(self) -> tuple[int, int, int]: ...

    @property
    @abc.abstractmethod
    def base3_address(self) -> str: ...

    @base3_address.setter
    @abc.abstractmethod
    def base3_address(self, value: str) -> None: ...

    @property
    @abc.abstractmethod
    def session_id(self) -> uuid.UUID: ...


class CommBufferSingleton(CommBuffer, Thread):
    def __init__(
        self,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        ser2: bool = False,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        if baudrate not in DEFAULT_VALID_BAUDRATES:
            raise ValueError(f"Invalid baudrate: {baudrate}")
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} TMCC Command Buffer")
        self._baudrate = baudrate
        self._port = port
        self._queue_size = queue_size
        self._ser2 = ser2
        if queue_size:
            self._queue = Queue(queue_size)
        else:
            self._queue = None
        self._base3_address = None
        self._shutdown_signalled = False
        self._last_output_at = 0  # used to throttle writes to LCS SER2
        # if there is no Ser2, send commands via Base 3
        from ..pdi.base3_buffer import Base3Buffer

        self._base3: Base3Buffer | None = None
        self._use_base3 = False
        self._tmcc_dispatcher = None
        self._uuid: uuid.UUID = uuid.uuid4()  # uniquely identify this instance of the server

        # start the consumer threads
        self._scheduler = DelayHandler(self)
        self.start()

    def update_state(self, state: ComponentState | CommandReq | PdiReq | bytes) -> None:
        """
        Force a state update through the system. Used to handle
        non-Lionel actors, like automatic train control blocks.
        """
        from ..pdi.constants import PDI_EOP, PDI_SOP
        from ..pdi.pdi_listener import PdiDispatcher
        from .command_listener import CommandDispatcher

        # if we got a state, remember it, otherwise, we have to
        # parse the byte stream and convert
        if isinstance(state, ComponentState):
            state = state.as_bytes
            if isinstance(state, list):
                state = b"".join(state)

        state_cmds = []
        if isinstance(state, bytes):
            # this is the hard one, the byte stream could be a mixture of PDI and TMCC
            # commands; we have to go thru and dispatch to the correct listener. With
            # that said, as a simplification, we will assume that the methods that call
            # this function are not mixed modal and the byte stream will either be a
            # CommandReq or a PdiReq
            # TODO: parse mixed modal stream
            if state[0] == PDI_SOP and state[-1] == PDI_EOP:
                state_cmds.append(PdiReq.from_bytes(state))
            else:
                state_cmds.append(CommandReq.from_bytes(state))
        elif isinstance(state, CommandReq):
            state_cmds.append(state)
        elif isinstance(state, PdiReq):
            state_cmds.append(state)
        else:
            raise AttributeError(f"Invalid state: {state}")
        for state_cmd in state_cmds:
            if isinstance(state_cmd, CommandReq):
                CommandDispatcher.get().offer(state_cmd)
            elif isinstance(state_cmd, PdiReq):
                PdiDispatcher.get().offer(state_cmd)

    def start_heart_beat(self, port: int = DEFAULT_SERVER_PORT):
        raise NotImplementedError

    @property
    def server_version(self) -> tuple[int, int, int]:
        from .. import get_version_tuple

        return get_version_tuple()

    @property
    def base3_address(self) -> str:
        return self._base3_address

    @base3_address.setter
    def base3_address(self, value: str) -> None:
        self._base3_address = value

    @property
    def session_id(self) -> uuid.UUID:
        return self._uuid

    @staticmethod
    def _current_milli_time() -> int:
        """
        Return the current time, in milliseconds past the "epoch"
        """
        return round(time.time() * 1000)

    @property
    def port(self) -> str:
        return self._port

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @property
    def is_ser2(self) -> bool:
        return self._ser2

    @property
    def is_use_base3(self) -> bool:
        return self._use_base3

    @is_use_base3.setter
    def is_use_base3(self, value: bool) -> None:
        from ..pdi.base3_buffer import Base3Buffer
        from .command_listener import CommandDispatcher

        self._use_base3 = value
        if value is True and self._base3 is None:
            with self._lock:
                if self._base3 is None:
                    self._base3 = Base3Buffer.get()
                if self._tmcc_dispatcher is None:
                    self._tmcc_dispatcher = CommandDispatcher.get()

    def enqueue_command(self, command: bytes, delay: float = 0) -> None:
        from ..pdi.constants import PDI_SOP

        if command:
            log.debug(f"Enqueue command 0x{command.hex()}")
            if delay > 0:
                self._scheduler.schedule(delay, command)
            else:
                if command[0] == PDI_SOP:
                    # send to Base 3 if one is available
                    if self._base3:
                        self._base3.send(command)
                    else:
                        log.error(f"Request to send packet to Base 3, but no Base 3 available: {command.hex()}")
                else:
                    self._queue.put(command)

    def shutdown(self, immediate: bool = False) -> None:
        if immediate:
            with self._queue.mutex:
                self._queue.queue.clear()
                self._queue.all_tasks_done.notify_all()
                self._queue.unfinished_tasks = 0
        self._shutdown_signalled = True

    def join(self) -> None:
        super().join()

    def register(self, port: int = DEFAULT_SERVER_PORT) -> None:
        pass  # noop; used to register client

    def disconnect(self, port: int = DEFAULT_SERVER_PORT) -> None:
        pass  # noop; used to disconnect client

    def sync_state(self, port: int = DEFAULT_SERVER_PORT) -> None:
        pass  # noop; used by client to request server state

    def run(self) -> None:
        # if the queue is not empty AND _shutdown_signaled is False, then exit
        while not self._queue.empty() or not self._shutdown_signalled:
            data = None
            try:
                data = self._queue.get(block=True, timeout=0.25)
                if self.is_use_base3 is True or self.is_ser2 is False:
                    self.base3_send(data)
                else:
                    self.ser2_send(data)
            except Empty:
                pass
            except Exception as e:
                log.error(f"Error sending {data.hex()}")
                log.exception(e)
            finally:
                if data is not None:
                    self._queue.task_done()

    def ser2_send(self, data):
        try:
            with serial.Serial(self.port, self.baudrate) as ser:
                millis_since_last_output = self._current_milli_time() - self._last_output_at
                if millis_since_last_output < DEFAULT_SER2_THROTTLE_DELAY:
                    time.sleep((DEFAULT_SER2_THROTTLE_DELAY - millis_since_last_output) / 1000.0)
                # Write the command byte sequence
                ser.write(data)
                self._last_output_at = self._current_milli_time()
                # inform Base 3 of state change, if available
                from ..pdi.base3_buffer import Base3Buffer

                Base3Buffer.sync_state(data)
        except SerialException as se:
            # TODO: handle serial errors
            log.exception(se)

    def base3_send(self, data: bytes):
        from ..pdi.constants import PdiCommand
        from ..pdi.pdi_req import TmccReq

        tmcc_cmd = CommandReq.from_bytes(data)
        if tmcc_cmd.address > 99:
            pdi_cmd = TmccReq(tmcc_cmd, PdiCommand.TMCC4_TX)
        else:
            pdi_cmd = TmccReq(tmcc_cmd, PdiCommand.TMCC_TX)
        self._base3.send(pdi_cmd.as_bytes)
        # also inform CommandDispatcher to update system state
        if self.is_ser2 is False or tmcc_cmd.is_force_state_update is True:
            self._tmcc_dispatcher.offer(tmcc_cmd)


COMM_ERROR_CODES: Dict[int, str] = {
    48: "ADDRESS IN USE",
    60: "TIMEOUT",
    61: "REFUSED",
    64: "HOST DOWN",
    65: "NO ROUTE",
    113: "NO ROUTE",
    120: "CONN ERR",
}


class CommBufferProxy(CommBuffer):
    """
    Allows a Raspberry Pi to "slave" to another so only one serial connection is needed
    """

    from ..db.component_state import ComponentState

    @classmethod
    def server_port(cls) -> int | None:
        # noinspection PyProtectedMember
        if cls.is_built() is True and cls.is_client() is True:
            # noinspection PyProtectedMember
            return cls._instance._port
        raise AttributeError("CommBufferProxy must be built first")

    # noinspection PyProtectedMember
    @classmethod
    def server_ip(cls) -> str:
        if cls.is_built() is True and cls.is_client() is True and cls._instance._ephemeral_port:
            return cls._instance._ephemeral_port[0]
        raise AttributeError("CommBufferProxy must be built first")

    def __init__(self, server: IPv4Address | IPv6Address = None, port: int = DEFAULT_SERVER_PORT) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        super().__init__()
        self._scheduler = DelayHandler(self)
        self._server = server
        self._port = port
        self._ephemeral_port = None
        self._client_port = None
        self._base3_address = None
        self._server_version = None
        self._uuid: uuid.UUID = uuid.uuid4()
        self._server_version_available = Event()
        self._heart_beat_thread = None

    def start_heart_beat(self, port: int = DEFAULT_SERVER_PORT):
        self._heart_beat_thread = ClientHeartBeat(self)

    @property
    def server_version(self) -> tuple[int, int, int]:
        return self._server_version

    @property
    def base3_address(self) -> str:
        return self._base3_address

    @base3_address.setter
    def base3_address(self, value: str) -> None:
        pass

    @property
    def session_id(self) -> uuid.UUID:
        return self._uuid

    @property
    def client_port(self) -> int:
        return self._client_port

    def server_version_available(self) -> Event:
        return self._server_version_available

    def enqueue_command(self, command: bytes, delay: float = 0) -> None:
        if delay > 0:
            self._scheduler.schedule(delay, command)
        else:
            retries = 0
            while True:
                try:
                    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                        s.settimeout(5.0)
                        s.connect((str(self._server), self._port))
                        s.settimeout(None)
                        s.sendall(command)
                        resp = s.recv(32)  # response contains Base 3 Addr as well as the server version
                        if self._server_version is None and len(resp) >= 3:
                            self._server_version = (resp[0], resp[1], resp[2])
                            self._server_version_available.set()
                        resp = resp[3:] if len(resp) > 3 else resp
                        if self._base3_address is None:
                            self._base3_address = resp.decode("utf-8", "ignore")
                        if self._ephemeral_port is None:
                            self._ephemeral_port = s.getsockname()
                    return
                except OSError as oe:
                    if retries < 90:
                        retries += 1
                        if retries % 5 == 0:
                            e_msg = COMM_ERROR_CODES.get(oe.errno, f"UNKNOWN ({oe.errno})")
                            log.info(f"Looking for {PROGRAM_NAME} server at {self._server}... [{e_msg}]")
                        if not isinstance(oe, TimeoutError):
                            time.sleep(1)
                        continue
                    raise oe

    def update_state(self, state: ComponentState | CommandReq | PdiReq | bytes) -> None:
        """
        Allow a state update to be sent to server and all clients.
        Implemented to support automatic train control Blocks
        but could support other IPC state updates that are not of
        Lionel origin in the future.
        """
        from .enqueue_proxy_requests import SENDING_STATE_REQUEST

        if state:
            if isinstance(state, ComponentState) or isinstance(state, CommandReq) or isinstance(state, PdiReq):
                state_bytes = state.as_bytes
            elif isinstance(state, bytes):
                state_bytes = state
            else:
                raise ValueError(f"Invalid state: {state}")
            if isinstance(state_bytes, bytes):
                state_bytes = [state_bytes]
            elif isinstance(state_bytes, list):
                pass
            else:
                raise ValueError(f"Invalid state bytes format: {state}: {type(state_bytes)}")
            for packet in state_bytes:
                self.enqueue_command(SENDING_STATE_REQUEST + packet)

    def register(self, port: int = DEFAULT_SERVER_PORT) -> None:
        from ..comm.enqueue_proxy_requests import EnqueueProxyRequests

        self._client_port = port
        while True:
            from .. import get_version_tuple

            self.enqueue_command(
                EnqueueProxyRequests.register_request(
                    port,
                    self.session_id,
                    get_version_tuple(),
                )
            )
            return

    def disconnect(self, port: int = None) -> None:
        port = self._client_port if port is None else port
        try:
            from ..comm.enqueue_proxy_requests import EnqueueProxyRequests

            if self._heart_beat_thread:
                self._heart_beat_thread.shutdown()
            self.enqueue_command(EnqueueProxyRequests.disconnect_request(port, self.session_id))
            return
        except ConnectionError as ce:
            raise ce

    def sync_state(self, port: int = None) -> None:
        port = self._client_port if port is None else port
        try:
            from ..comm.enqueue_proxy_requests import EnqueueProxyRequests

            # noinspection PyTypeChecker
            self.enqueue_command(EnqueueProxyRequests.sync_state_request(port, self.session_id))
        except ConnectionError as ce:
            raise ce

    def shutdown(self, immediate: bool = False) -> None:
        pass

    def join(self):
        pass


class DelayHandler(Thread):
    """
    Handle delayed (scheduled) requests. Implementation uses Python's lightweight
    sched module to keep a list of requests to issue in the future. We use
    threading.Event.wait() as the sleep function, as it is interruptable. This
    allows us to schedule requests in any order and still have them fire at the
    appropriate time.
    """

    def __init__(self, buffer: CommBuffer) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Command Delay Handler")
        self._buffer = buffer
        self._cv = Condition()
        self._ev = Event()
        self._scheduler = sched.scheduler(time.time, self._ev.wait)
        self.start()

    def run(self) -> None:
        while True:
            with self._cv:
                while self._scheduler.empty():
                    self._cv.wait()
            # run the scheduler outside the cv lock; otherwise,
            # we couldn't schedule more commands
            self._scheduler.run()
            self._ev.clear()

    def schedule(self, delay: float, command: bytes) -> None:
        with self._cv:
            self._scheduler.enter(delay, 1, self._buffer.enqueue_command, (command,))
            # this interrupts the running scheduler
            self._ev.set()
            # and this notifies the main thread to restart, as there is a new
            # request in the sched queue
            self._cv.notify()


class ClientHeartBeat(Thread):
    def __init__(self, tmcc_buffer: CommBufferProxy) -> None:
        from .. import CommandReq

        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Client Heart Beat")
        self._tmcc_buffer = tmcc_buffer
        self._client_port = tmcc_buffer.client_port
        heartbeat = CommandReq(TMCC1SyncCommandEnum.KEEP_ALIVE).as_bytes
        self._heartbeat_bytes = (
            heartbeat + int(self._client_port & 0xFFFF).to_bytes(2, byteorder="big") + tmcc_buffer.session_id.bytes
        )
        self._ev = threading.Event()
        self._is_running = True
        self.start()

    def shutdown(self) -> None:
        self._is_running = False
        self._ev.set()

    def run(self) -> None:
        while self._is_running:
            self._ev.wait(DEFAULT_PULSE)
            if self._is_running:
                self._tmcc_buffer.enqueue_command(self._heartbeat_bytes)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
import time
from threading import Condition, Event, Thread

from ..comm.command_listener import SYNC_COMPLETE, SYNCING, CommandDispatcher
from ..db.component_state_store import ComponentStateStore
from ..pdi.amc2_req import Amc2StateSync
from ..pdi.base_req import BaseReq
from ..pdi.constants import D4Action, PdiCommand
from ..pdi.d4_req import D4Req
from ..pdi.pdi_listener import PdiListener
from ..pdi.pdi_req import AllReq, PdiReq
from ..pdi.pdi_state_store import PdiStateStore
from ..protocol.constants import PROGRAM_NAME, CommandScope

log = logging.getLogger(__name__)


class StartupState(Thread):
    def __init__(
        self,
        listener: PdiListener,
        dispatcher: CommandDispatcher,
        pdi_state_store: PdiStateStore,
    ) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Startup State Sniffer")
        self.pdi_listener = listener
        self.pdi_dispatcher = listener.dispatcher if listener else None
        self.pdi_state_store = pdi_state_store
        self._cv = Condition()
        self._ev = Event()
        self._waiting_for = dict()
        self._processed_configs = set()
        self._sync_state = ComponentStateStore.get_state(CommandScope.SYNC, 99)
        self._dispatcher = dispatcher
        self._dispatcher.offer(SYNCING)
        self.start()

    def __call__(self, cmd: PdiReq) -> None:
        """
        Callback specified in the Subscriber protocol used to send events to listeners
        """

        if cmd:
            with self._cv:
                self._waiting_for.pop(cmd.as_key, None)
        else:
            return
        if isinstance(cmd, PdiReq):
            req = None
            if cmd.action and cmd.action.is_config and self._config_key(cmd) not in self._processed_configs:
                # register the device; registration returns a list of pdi commands
                # to send to get device state
                state_requests = self.pdi_state_store.register_pdi_device(cmd)
                self._processed_configs.add(self._config_key(cmd))
                if state_requests:
                    for state_request in state_requests:
                        self.pdi_listener.enqueue_command(state_request)
            elif isinstance(cmd, BaseReq) and cmd.pdi_command == PdiCommand.BASE_MEMORY:
                if cmd.scope == CommandScope.TRAIN and cmd.tmcc_id == 98:
                    self._dispatcher.offer(SYNC_COMPLETE)
                # send a request to the base to get the next engine/train/acc/switch/route record (0x26)
                if cmd.tmcc_id < 98 and cmd.data_length == PdiReq.scope_record_length(cmd.scope):
                    req = BaseReq(cmd.tmcc_id + 1, PdiCommand.BASE_MEMORY, scope=cmd.scope)
            elif isinstance(cmd, D4Req):
                req = None
                if cmd.action == D4Action.COUNT and cmd.count:
                    # request first record of D4 engines/trains
                    req = D4Req(0, cmd.pdi_command, D4Action.FIRST_REC)
                elif cmd.action in {D4Action.FIRST_REC, D4Action.NEXT_REC}:
                    if cmd.next_record_no == 0xFFFF:
                        pass
                    elif cmd.next_record_no is not None:
                        # query current state of 4-digit engine/train
                        req = D4Req(
                            cmd.next_record_no,
                            cmd.pdi_command,
                            D4Action.QUERY,
                            start=0,
                            data_length=0xC0,
                        )
                        with self._cv:
                            self._waiting_for[req.as_key] = req
                        self.pdi_listener.enqueue_command(req)
                        # get the record number of the next engine/train
                        req = D4Req(cmd.next_record_no, cmd.pdi_command, D4Action.NEXT_REC)
            if req:
                with self._cv:
                    self._waiting_for[req.as_key] = req
                self.pdi_listener.enqueue_command(req)
            else:
                with self._cv:
                    if not self._waiting_for:
                        self._ev.set()
                        self._cv.notify_all()

    @staticmethod
    def _config_key(cmd: PdiReq) -> bytes:
        """
        Generates a unique configuration key based on the provided PdiReq command.

        Summary:
        This static method combines the byte representations of various attributes
        from a PdiReq command object to generate a unique byte key. The resulting
        key is composed of the PdiReq command's bytes, tmcc_id converted to a single
        byte, and the action's bytes.

        Args:
        cmd (PdiReq): The command object containing the attributes to be used in
        constructing the configuration key.

        Returns:
        bytes: The concatenated byte string representing the configuration key.
        """
        byte_str = cmd.pdi_command.as_bytes
        byte_str += cmd.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += cmd.action.as_bytes
        return byte_str

    def run(self) -> None:
        self.pdi_listener.subscribe_any(self)
        self.pdi_listener.enqueue_command(AllReq())
        self.pdi_listener.enqueue_command(BaseReq(0, PdiCommand.BASE))
        for pdi_command in [PdiCommand.D4_ENGINE, PdiCommand.D4_TRAIN]:
            req = D4Req(0, pdi_command, D4Action.COUNT)
            with self._cv:
                self._waiting_for[req.as_key] = req
            self.pdi_listener.enqueue_command(req)
        # Request engine/sw/acc roster at startup; do this by asking for
        # Eng/Train/Acc/Sw/Route then examining the rev links returned until
        # we find one out of range; make a request for each discovered entity
        for scope in [
            CommandScope.ENGINE,
            CommandScope.TRAIN,
            CommandScope.SWITCH,
            CommandScope.ACC,
            CommandScope.ROUTE,
        ]:
            req = BaseReq(1, PdiCommand.BASE_MEMORY, scope=scope)
            with self._cv:
                self._waiting_for[req.as_key] = req
            self.pdi_listener.enqueue_command(req)

        # now wait for all responses; this will not track LCS devices reporting their config
        # because of the AllReq
        total_time = 0
        now = time.time()
        ev_set = False
        while total_time < 120:  # only listen for 2 minutes
            self._ev.wait(0.25)
            if self._ev.is_set() or (ev_set is True):
                self._ev.clear()
                ev_set = True
                if round(time.time() - now) >= 0:
                    log.info(f"Initial state loaded from Base 3: {time.time() - now:.2f} seconds elapsed.")
                    break
            total_time += 0.25
        # register handlers for special cases where additional processing
        # is required to process state
        _ = Amc2StateSync(self.pdi_listener)
        # print out any stragglers; this is an error we should address
        for k, v in self._waiting_for.items():
            log.info(f"Failed to receive {k} state: {v}")

        self.pdi_listener.unsubscribe_any(self)



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

#
import logging
from argparse import ArgumentParser
from typing import List

from ..protocol.tmcc1.switch_cmd import SwitchCmd
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum
from ..utils.argument_parser import PyTrainArgumentParser
from . import CliBase

log = logging.getLogger(__name__)


class SwitchCli(CliBase):
    @classmethod
    def command_parser(cls) -> ArgumentParser:
        sw_parser = PyTrainArgumentParser(add_help=False)
        sw_parser.add_argument("switch", metavar="Switch Number", type=int, help="switch to fire")
        group = sw_parser.add_mutually_exclusive_group()
        group.add_argument(
            "-through",
            action="store_const",
            const=TMCC1SwitchCommandEnum.THRU,
            dest="command",
            help="Throw Through",
        )
        group.add_argument(
            "-out", action="store_const", const=TMCC1SwitchCommandEnum.OUT, dest="command", help="Throw Out"
        )
        group.add_argument(
            "-address",
            action="store_const",
            const=TMCC1SwitchCommandEnum.SET_ADDRESS,
            dest="command",
            help="Set switch address",
        )
        group.set_defaults(command=TMCC1SwitchCommandEnum.THRU)
        return PyTrainArgumentParser("Throw specified switch (1 - 99)", parents=[sw_parser, cls.cli_parser()])

    """
        Throw the specified switch.

        Currently only available via the TMCC1 command format
    """

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        self._switch = self._args.switch
        self._switch_state = self._args.command
        try:
            cmd = SwitchCmd(
                self._switch, self._switch_state, baudrate=self._baudrate, port=self._port, server=self._server
            )
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



from __future__ import annotations


import sys

from .multibyte_command_req import MultiByteReq
from .multibyte_constants import TMCC2R4LCIndex, TMCC2R4LCEnum, UnitAssignment

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from ..constants import DEFAULT_ADDRESS, CommandScope
from ..tmcc2.tmcc2_constants import LEGACY_MULTIBYTE_COMMAND_PREFIX
from ..tmcc2.tmcc2_constants import LEGACY_TRAIN_COMMAND_PREFIX

"""
Commands to modify R4LC EEPROM
"""


class R4LCCommandReq(MultiByteReq):
    @classmethod
    def build(
        cls, command: TMCC2R4LCEnum, address: int = DEFAULT_ADDRESS, data: int = 0, scope: CommandScope = None
    ) -> Self:
        return R4LCCommandReq(command, address, data, scope)

    @classmethod
    def from_bytes(cls, param: bytes, from_tmcc_rx: bool = False, is_tmcc4: bool = False) -> Self:
        if not param:
            raise ValueError("Command requires at least 9 bytes")
        if len(param) < 9:
            raise ValueError(f"R4LC command requires at least 9 bytes {param.hex(':')}")
        if (
            len(param) == 9
            and param[3] == LEGACY_MULTIBYTE_COMMAND_PREFIX
            and param[6] == LEGACY_MULTIBYTE_COMMAND_PREFIX
        ):
            index = 0x00FF & int.from_bytes(param[1:3], byteorder="big")
            try:
                pi = TMCC2R4LCIndex(index)
            except ValueError:
                raise ValueError(f"Invalid R4LC command: : {param.hex(':')}")
            cmd_enum = TMCC2R4LCEnum(pi.name)
            data = int(param[5])
            scope = CommandScope.ENGINE
            if int(param[0]) == LEGACY_TRAIN_COMMAND_PREFIX:
                scope = CommandScope.TRAIN
            # build_req the request and return
            address = cmd_enum.value.address_from_bytes(param[1:3])
            cmd_req = R4LCCommandReq.build(cmd_enum, address, data, scope)
            if from_tmcc_rx is True:
                cmd_req._is_tmcc_rx = True
            return cmd_req
        raise ValueError(f"Invalid R4LC command: : {param.hex(':')}")

    def __init__(
        self,
        command_def_enum: TMCC2R4LCEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
    ) -> None:
        super().__init__(command_def_enum, address, data, scope)

    def __repr__(self) -> str:
        if self.command == TMCC2R4LCEnum.TRAIN_UNIT:
            data = UnitAssignment.by_value(self._data)
            if data is not None:
                return f"[{self.scope.name} {self.address} {self.command_name} {data.name} (0x{self.as_bytes.hex()})]"
        return super().__repr__()

    @property
    def index_byte(self) -> bytes:
        return self.command_def.bits.to_bytes(1, byteorder="big")

    @property
    def data_byte(self) -> bytes:
        return (0x00).to_bytes(1, byteorder="big")
